#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)
import logging
from tabulate import tabulate
from typing import List

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from fabrictestbed_extensions.fablib.slice import Slice
    from fabrictestbed_extensions.fablib.interface import Interface
    from fabric_cf.orchestrator.swagger_client import Sliver as OrchestratorSliver

from fabrictestbed.slice_editor import ServiceType, NetworkService as FimNetworkService
from fim.slivers.network_service import ServiceType, NSLayer

from ipaddress import IPv4Address, IPv6Address, IPv4Network, IPv6Network


class NetworkService:
    network_service_map = { 'L2Bridge': ServiceType.L2Bridge,
                            'L2PTP': ServiceType.L2PTP,
                            'L2STS': ServiceType.L2STS,
                            }

    # Type names used in fim network services
    fim_l2network_service_types = [ 'L2Bridge', 'L2PTP', 'L2STS']
    fim_l3network_service_types = [ 'FABNetv4', 'FABNetv6']


    @staticmethod
    def get_fim_l2network_service_types() -> List[str]:
        """
        Not inteded for API use
        """
        return NetworkService.fim_l2network_service_types

    @staticmethod
    def get_fim_l3network_service_types() -> List[str]:
        """
        Not inteded for API use
        """
        return NetworkService.fim_l3network_service_types

    @staticmethod
    def get_fim_network_service_types() -> List[str]:
        """
        Not inteded for API use
        """
        return NetworkService.get_fim_l2network_service_types() + NetworkService.get_fim_l3network_service_types()

    @staticmethod
    def calculate_l2_nstype(interfaces: List[Interface] = None) -> ServiceType:
        """
        Not inteded for API use

        Determines the L2 network service type based on the number of interfaces inputted.

        :param interfaces: a list of interfaces
        :type interfaces: list[Interface]
        :raises Exception: if no network service type is not appropriate for the number of interfaces
        :return: the network service type
        :rtype: ServiceType
        """
        # if there is a basic NIC, WAN must be STS
        basic_nic_count = 0

        sites = set([])
        for interface in interfaces:
            sites.add(interface.get_site())
            if interface.get_model() == "NIC_Basic":
                basic_nic_count += 1

        rtn_nstype = None
        if len(sites) == 1:
            rtn_nstype = NetworkService.network_service_map['L2Bridge']
        #elif basic_nic_count == 0 and len(sites) == 2 and len(interfaces) == 2:
        #    #TODO: remove this when STS works on all links.
        #    rtn_nstype = NetworkService.network_service_map['L2PTP']
        elif len(sites) == 2:
            if len(interfaces) >= 2:
                rtn_nstype = NetworkService.network_service_map['L2STS']
        else:
            raise Exception(f"Invalid Network Service: Networks are limited to 2 unique sites. Site requested: {sites}")

        return rtn_nstype

    @staticmethod
    def validate_nstype(type: ServiceType, interfaces: List[Interface]) -> bool:
        """
        Not inteded for API use


        Verifies the network service type against the number of interfaces.

        :param type: the network service type to check
        :type type: ServiceType
        :param interfaces: the list of interfaces to check
        :type interfaces: list[Interface]
        :raises Exception: if the network service type is invalid based on the number of interfaces
        :return: true if the network service type is valid based on the number of interfaces
        :rtype: bool
        """
        sites = set([])
        nics = set([])
        nodes = set([])
        for interface in interfaces:
            sites.add(interface.get_site())
            nics.add(interface.get_model())
            nodes.add(interface.get_node())

        # models: 'NIC_Basic', 'NIC_ConnectX_6', 'NIC_ConnectX_5'
        if type == NetworkService.network_service_map['L2Bridge']:
            if not len(sites) == 1:
                raise Exception(f"Network type {type} must include interfaces from exactly one site. "
                                f"{len(sites)} sites requested: {sites}")

        elif type == NetworkService.network_service_map['L2PTP']:
            if not len(sites) == 2:
                raise Exception(f"Network type {type} must include interfaces from exactly two sites. "
                                f"{len(sites)} sites requested: {sites}")
            if 'NIC_Basic' in nics:
                raise Exception(f"Network type {type} does not support interfaces of type 'NIC_Basic'")

        elif type == NetworkService.network_service_map['L2STS']:
            exception_list = []
            if len(sites) != 2:
                exception_list.append(f"Network type {type} must include interfaces from exactly two sites. "
                                      f"{len(sites)} sites requested: {sites}")
            if len(interfaces) > 2:
                hosts = set([])
                for interface in interfaces:
                    node = interface.get_node()
                    if interface.get_model() == 'NIC_Basic':
                        if node.get_host() is None:
                            exception_list.append(f"Network type {type} does not support multiple NIC_Basic "
                                                  f"interfaces on VMs residing on the same host. Please see "
                                                  f"Node.set_host(host_nane) to explicitily bind a nodes to a "
                                                  f"specific host. Node {node.get_name()} is unbound.")
                        elif node.get_host() in hosts:
                            exception_list.append(f"Network type {type} does not support multiple NIC_Basic interfaces "
                                                  f"on VMs residing on the same host. Please see "
                                                  f"Node.set_host(host_nane) to explicitily bind a nodes to a specific "
                                                  f"host. Multiple nodes bound to {node.get_host()}.")
                        else:
                            hosts.add(node.get_host())

            if len(exception_list) > 0:
                raise Exception(f"{exception_list}")
        else:
            raise Exception(f"Unknown network type {type}")

        return True

    @staticmethod
    def new_l3network(slice: Slice = None, name: str = None, interfaces: List[Interface] = [], type: str = None):
        """
        Not inteded for API use. See slice.add_l3network
        """
        if type == "IPv6":
            nstype = ServiceType.FABNetv6
        else:
            nstype = ServiceType.FABNetv4

        # TODO: need a fabnet version of this
        # validate nstype and interface List
        #NetworkService.validate_nstype(nstype, interfaces)

        return NetworkService.new_network_service(slice=slice, name=name, nstype=nstype, interfaces=interfaces)

    @staticmethod
    def new_l2network(slice: Slice = None, name: str = None, interfaces: List[Interface] = [], type: str = None):
        """
        Not inteded for API use. See slice.add_l2network

        Creates a new L2 network service.

        :param slice: the fablib slice to build this network on
        :type slice: Slice
        :param name: the name of the new network
        :type name: str
        :param interfaces: a list of interfaces to build the network service on
        :type interfaces: list[Interface]
        :param type: the type of network service to build (optional)
        :tyep type: str
        :return: the new L2 network service
        :rtype: NetworkService
        """
        if type is None:
            nstype = NetworkService.calculate_l2_nstype(interfaces=interfaces)
        else:
            if type in NetworkService.get_fim_l2network_service_types():
                nstype = NetworkService.network_service_map[type]
            else:
                raise Exception(f"Invalid l2 network type: {type}. Please choose from "
                                f"{NetworkService.get_fim_l2network_service_types()} or None for automatic selection")

        # validate nstype and interface List
        NetworkService.validate_nstype(nstype, interfaces)

        # Set default VLANs for P2P networks that did not assing VLANs
        if nstype is ServiceType.L2PTP: # or nstype == ServiceType.L2STS:
            for interface in interfaces:
                if interface.get_model() != 'NIC_Basic' and not interface.get_vlan():
                    # TODO: Long term we might have muliple vlan on one property
                    # and will need to make sure they are unique.  For now this okay
                    interface.set_vlan("100")

        return NetworkService.new_network_service(slice=slice, name=name, nstype=nstype, interfaces=interfaces)

    @staticmethod
    def new_network_service(slice: Slice = None, name: str = None, nstype: ServiceType = None,
                            interfaces: List[Interface] = []):
        """
        Not inteded for API use. See slice.add_l2network


        Creates a new FABRIC network service and returns the fablib instance.

        :param slice: the fabric slice to build the network service with
        :type slice: Slice
        :param name: the name of the new network service
        :type name: str
        :param nstype: the type of network service to create
        :type nstype: ServiceType
        :param interfaces: a list of interfaces to
        :return: the new fablib network service
        :rtype: NetworkService
        """
        fim_interfaces = []
        for interface in interfaces:
            fim_interfaces.append(interface.get_fim_interface())

        logging.info(f"Create Network Service: Slice: {slice.get_name()}, Network Name: {name}, Type: {nstype}")
        fim_network_service = slice.topology.add_network_service(name=name,
                                                                 nstype=nstype,
                                                                 interfaces=fim_interfaces)

        return NetworkService(slice=slice, fim_network_service=fim_network_service)

    @staticmethod
    def get_l3network_services(slice: Slice = None) -> list:
        """
        Not inteded for API use.
        """
        topology = slice.get_fim_topology()

        rtn_network_services = []
        fim_network_service = None
        logging.debug(f"NetworkService.get_fim_l3network_service_types(): {NetworkService.get_fim_l3network_service_types()}")

        for net_name, net in topology.network_services.items():
            logging.debug(f"scanning network: {net_name}, net: {net}")
            if str(net.get_property('type')) in NetworkService.get_fim_l3network_service_types():
                logging.debug(f"returning network: {net_name}, net: {net}")
                rtn_network_services.append(NetworkService(slice = slice, fim_network_service = net))

        return rtn_network_services

    @staticmethod
    def get_l3network_service(slice: Slice = None, name: str = None):
        """
        Not inteded for API use.
        """
        for net in NetworkService.get_l3network_services(slice=slice):
            if net.get_name() == name:
                return net

        raise Exception(f"Network not found. Slice {slice.slice_name}, network {name}")

    @staticmethod
    def get_l2network_services(slice: Slice = None) -> list:
        """
        Not inteded for API use.

        Gets a list of L2 network services on a fablib slice.

        :param slice: the fablib slice from which to get the network services
        :type slice: Slice
        :return: a list of network services on slice
        :rtype: list[NetworkService]
        """
        topology = slice.get_fim_topology()

        rtn_network_services = []
        fim_network_service = None
        for net_name, net in topology.network_services.items():
            if str(net.get_property('type')) in NetworkService.get_fim_l2network_service_types():
                rtn_network_services.append(NetworkService(slice=slice, fim_network_service=net))

        return rtn_network_services

    @staticmethod
    def get_l2network_service(slice: Slice = None, name: str = None):
        """
        Not inteded for API use.

        Gets a particular network service on a fablib slice.

        :param slice: the fablib slice from which to get the network service
        :type slice: Slice
        :param name: the name of the network service to get
        :type name: str
        :raises Exception: if the network is not found
        :return: the particular network service
        :rtype: NetworkService
        """
        for net in NetworkService.get_l2network_services(slice=slice):
            if net.get_name() == name:
                return net

        raise Exception(f"Network not found. Slice {slice.slice_name}, network {name}")

    @staticmethod
    def get_network_services(slice: Slice = None) -> list:
        """
        Not inteded for API use.
        """

        topology = slice.get_fim_topology()

        rtn_network_services = []
        fim_network_service = None
        for net_name, net in topology.network_services.items():
            if str(net.get_property('type')) in NetworkService.get_fim_network_service_types():
                rtn_network_services.append(NetworkService(slice=slice, fim_network_service = net))

        return rtn_network_services

    @staticmethod
    def get_network_service(slice: Slice = None, name: str = None):
        """
        Not inteded for API use.
        """
        for net in NetworkService.get_network_services(slice=slice):
            if net.get_name() == name:
                return net

        raise Exception(f"Network not found. Slice {slice.slice_name}, network {name}")

    def __init__(self, slice: Slice = None, fim_network_service: FimNetworkService =None):
        """
        Not inteded for API use.

        Constructor. Sets the fablib slice and the FABRIC network service.

        :param slice: the fablib slice to set as instance state
        :type slice: Slice
        :param fim_network_service: the FIM network service to set as instance state
        :type fim_network_service: FIMNetworkService
        """
        super().__init__()
        self.fim_network_service = fim_network_service
        self.slice = slice

        try:
            self.sliver = slice.get_sliver(reservation_id=self.get_reservation_id())
        except:
            self.sliver = None

    def __str__(self):
        """
        Creates a tabulated string describing the properties of the network service.

        Intended for printing network service information.

        :return: Tabulated string of network service information
        :rtype: String
        """
        table = [ ["ID", self.get_reservation_id()],
            ["Name", self.get_name()],
            ["Layer", self.get_layer()],
            ["Type", self.get_type()],
            ["Site", self.get_site()],
            ["Gateway", self.get_gateway()],
            ["L3 Subnet", self.get_subnet()],
            ["Reservation State", self.get_reservation_state()],
            ["Error Message", self.get_error_message()],
            ]

        return tabulate(table) #, headers=["Property", "Value"])

    def get_slice(self) -> Slice:
        """
        Gets the fablib slice this network service is built on.

        :return: the slice this network is on
        :rtype: Slice
        """
        return self.slice

    def get_site(self) -> str or None:
        try:
            return self.get_sliver().sliver.site
        except Exception as e:
            logging.warning(f"Failed to get site: {e}")

            return None

    def get_layer(self) -> str or None:
        """
        Gets the layer of the network services (L2 or L3)

        :return: L2 or L3
        :rtype: String
        """
        try:
            return self.get_sliver().sliver.layer
        except Exception as e:
            logging.warning(f"Failed to get layer: {e}")
            return None

    def get_type(self):
        """
        Gets the type of the network services

        :return: network service types
        :rtype: String
        """
        try:
            return self.get_sliver().sliver.resource_type
        except Exception as e:
            logging.warning(f"Failed to get type: {e}")
            return None

    def get_sliver(self) -> OrchestratorSliver:
        return self.sliver

    def get_fim_network_service(self) -> FimNetworkService:
        """
        Not intended for API use

        Gets the FABRIC network service this instance represents.

        :return: the FIM network service
        :rtype: FIMNetworkService
        """
        return self.fim_network_service

    def get_error_message(self) -> str:
        """
        Gets the error messages

        :return: network service types
        :rtype: String
        """
        try:
            return self.get_fim_network_service().get_property(pname='reservation_info').error_message
        except:
            return ""

    def get_gateway(self) -> IPv4Address or IPv6Address or None:
        """
        Gets the assigend gateway for a FABnetv L3 IPv6 or IPv4 network

        :return: gateway IP
        :rtype: IPv4Address or IPv6Network
        """
        try:
            gateway = None
            if self.get_layer() == NSLayer.L3:
                if self.get_type() == ServiceType.FABNetv6:
                    gateway = IPv6Address(self.get_sliver().sliver.gateway.gateway)
                elif self.get_type() == ServiceType.FABNetv4:
                    gateway = IPv4Address(self.get_sliver().sliver.gateway.gateway)

            return gateway
        except Exception as e:
            logging.warning(f"Failed to get gateway: {e}")
            return None

    def get_available_ips(self, count: int = 256) -> List[IPv4Address or IPv6Address] or None:
        """
        Gets the IPs available for a FABnet L3 network.

        Note: large IPv6 address spaces take considerable time to build this
        list. By default this will return the first 256 addresses.  If you needed
        more addresses, set the count parameter.

        :param count: number of addresse to include
        :type slice: Slice
        :return: gateway IP
        :rtype: List[IPv4Address]
        """
        try:
            ip_list = []
            gateway = self.get_gateway()
            for i in range(count):
                logging.debug(f"adding IP {i}")
                ip_list.append(gateway+i+1)
            return ip_list
        except Exception as e:
            logging.warning(f"Failed to get_available_ips: {e}")
            return None

    def get_subnet(self) -> IPv4Network or IPv6Network or None:
        """
        Gets the assigend subnet for a FABnetv L3 IPv6 or IPv4 network

        :return: gateway IP
        :rtype: IPv4Network or IPv6Network
        """
        try:
            subnet = None
            if self.get_layer() == NSLayer.L3:
                if self.get_type() == ServiceType.FABNetv6:
                    subnet = IPv6Network(self.get_sliver().sliver.gateway.subnet)
                elif self.get_type() == ServiceType.FABNetv4:
                    subnet = IPv4Network(self.get_sliver().sliver.gateway.subnet)
            return subnet
        except Exception as e:
            logging.warning(f"Failed to get subnet: {e}")
            return None

    def get_reservation_id(self) -> str or None:
        """
        Gets the reservation id of the network

        :return: reservation ID
        :rtype: String
        """
        try:
            return self.get_fim_network_service().get_property(pname='reservation_info').reservation_id
        except Exception as e:
            logging.warning(f"Failed to get reservation_id: {e}")
            return None

    def get_reservation_state(self) -> str or None:
        """
        Gets the reservation state of the network

        :return: reservation state
        :rtype: String
        """
        try:
            return self.get_fim_network_service().get_property(pname='reservation_info').reservation_state
        except Exception as e:
            logging.warning(f"Failed to get reservation_state: {e}")
            return None

    def get_name(self) -> str:
        """
        Gets the name of this network service.

        :return: the name of this network service
        :rtype: String
        """
        return self.get_fim_network_service().name

    def get_interfaces(self) -> List[Interface]:
        """
        Gets the interfaces on this network service.

        :return: the interfaces on this network service
        :rtype: List[Interfaces]
        """
        interfaces = []
        for interface in self.get_fim_network_service().interface_list:
            interfaces.append(self.get_slice().get_interface(name=interface.name))

        return interfaces

    def get_interface(self, name: str = None) -> Interface or None:
        """
        Gets a particular interface on this network service.

        :param name: the name of the interface to search for
        :type name: str
        :return: the particular interface
        :rtype: Interface
        """
        for interface in self.get_fim_network_service().interface_list:
            if name in interface.name:
                return self.get_slice().get_interface(name=interface.name)

        return None

    def has_interface(self, interface: Interface) -> bool:
        """
        Determines whether this network service has a particular interface.

        :param interface: the fablib interface to search for
        :type interface: Interface
        :return: whether this network service has interface
        :rtype: bool
        """
        for fim_interface in self.get_fim_network_service().interface_list:
            #print(f"fim_interface.name: {fim_interface.name}, interface.get_name(): {interface.get_name()}")
            if fim_interface.name.endswith(interface.get_name()):
                return True

        return False



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

import os
import traceback
import re

import functools
import time

import importlib.resources as pkg_resources
from typing import List

from fabrictestbed.slice_editor import Labels, ExperimentTopology, Capacities, ComponentType, ComponentModelType, ServiceType, ComponentCatalog
from fabrictestbed.slice_editor import (
    ExperimentTopology,
    Capacities
)
from fabrictestbed.slice_manager import SliceManager, Status, SliceState


from .abc_test import AbcTest

from .. import images


class GPUTeslaT4Benchmark(AbcTest):

    def __init__(self):
        """
        Constructor
        :return:
        """
        super().__init__()




    def create_slice(self, slice_name, site_name, node_count=1):
        image = 'default_centos_8'
        image_type = 'qcow2'
        cores = 4
        ram = 8
        disk = 50
        model_type = ComponentModelType.GPU_Tesla_T4
        #Tesla_T4_component_type = ComponentType.Tesla_T4
        #Tesla_T4_model = 'P4510'
        Tesla_T4_name = 'gpu1'

        self.topology = ExperimentTopology()

        cap = Capacities(core=cores, ram=ram, disk=disk)

        for node_num in range(0, node_count):
            node_name="Tesla_T4-"+str(site_name)+"-"+str(node_num)
            # Add node
            n1 = self.topology.add_node(name=node_name, site=site_name)

            # Set Properties
            n1.set_properties(capacities=cap, image_type=image_type, image_ref=image)

            # Add the PCI Tesla_T4 device
            n1.add_component(model_type=model_type, name=Tesla_T4_name)
            #n1.add_component(ctype=Tesla_T4_component_type, model=Tesla_T4_model, name=Tesla_T4_name)

        # Generate Slice Graph
        slice_graph = self.topology.serialize()

        # Request slice from Orchestrator
        return_status, slice_reservations = self.slice_manager.create(slice_name=slice_name,
                                                    slice_graph=slice_graph,
                                                    ssh_key=self.node_ssh_key)




        if return_status == Status.OK:
            slice_id = slice_reservations[0].get_slice_id()
            #print("Submitted slice creation request. Slice ID: {}".format(slice_id))
        else:
            print(f"Failure: {slice_reservations}")

        #time.sleep(30)

        return_status, slices = self.slice_manager.slices(excludes=[SliceState.Dead,SliceState.Closing])

        if return_status == Status.OK:
            self.slice = list(filter(lambda x: x.slice_name == slice_name, slices))[0]
            #self.slice = self.wait_for_slice(slice, progress=True)

        #print()
        #print("Slice Name : {}".format(self.slice.slice_name))
        #print("ID         : {}".format(self.slice.slice_id))
        #print("State      : {}".format(self.slice.slice_state))
        #print("Lease End  : {}".format(self.slice.lease_end))

    def config_test(self):
        for node_name, node in self.topology.nodes.items():
            #print("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
            #print("Node:")
            print("Config | ", end = '')
            print(str(node_name) + " | ", end = '')
            try:
                print("{}".format(str(node_name)) + " | ", end = '')

                #print("   Name              : {}".format(node_name))
                #print("   Cores             : {}".format(node.get_property(pname='capacity_allocations').core))
                #print("   RAM               : {}".format(node.get_property(pname='capacity_allocations').ram))
                #print("   Disk              : {}".format(node.get_property(pname='capacity_allocations').disk))
                #print("   Image             : {}".format(node.image_ref))
                #print("   Image Type        : {}".format(node.image_type))
                #print("   Host              : {}".format(node.get_property(pname='label_allocations').instance_parent))
                print("{}".format(str(node.get_property(pname='label_allocations').instance_parent)).replace('\n','') + " | ", end = '')
                #print("   Site              : {}".format(node.site))
                print("{}".format(str(node.management_ip)).replace('\n','') + " | ", end = '')

                #print("   Management IP     : {}".format(node.management_ip))
                #print("   Reservation ID    : {}".format(node.get_property(pname='reservation_info').reservation_id))
                #print("   Reservation State : {}".format(node.get_property(pname='reservation_info').reservation_state))
                #print("   Components        : {}".format(node.components))
                #print("   Interfaces        : {}".format(node.interfaces))
                #print()
                name = node_name
                management_ip = node.management_ip

                #print("------------------------- Test Output ---------------------------")
                script= '#!/bin/bash  \n' \
                'sudo yum install -y pciutils \n'
                try:
                    stdout_str = self.execute_script(node_username='centos', node=node, script=script)
                    print("Done ")
                except Exception as e:
                    print("Fail {}".format(str(e)))
                #print(str(stdout_str.replace('\n','')) + " | ", end = '')
                #print("-----------------------------------------------------------------")
                #stdout_str = str(stdout.read(),'utf-8').replace('\\n','\n')
            except Exception as e:
                print ("Error in test: Error {}".format(e))
                traceback.print_exc()

    def run_test(self):
        self.config_test()

        for node_name, node in self.topology.nodes.items():
            #print("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
            #print("Node:")
            print(str(node_name) + " | ", end = '')
            try:
                #print("   Name              : {}".format(node_name))
                #print("   Cores             : {}".format(node.get_property(pname='capacity_allocations').core))
                #print("   RAM               : {}".format(node.get_property(pname='capacity_allocations').ram))
                #print("   Disk              : {}".format(node.get_property(pname='capacity_allocations').disk))
                #print("   Image             : {}".format(node.image_ref))
                #print("   Image Type        : {}".format(node.image_type))
                #print("   Host              : {}".format(node.get_property(pname='label_allocations').instance_parent))
                #print("   Site              : {}".format(node.site))
                #print("   Management IP     : {}".format(node.management_ip))
                #print("   Reservation ID    : {}".format(node.get_property(pname='reservation_info').reservation_id))
                #print("   Reservation State : {}".format(node.get_property(pname='reservation_info').reservation_state))
                #print("   Components        : {}".format(node.components))
                #print("   Interfaces        : {}".format(node.interfaces))
                #print()
                name = node_name
                reservation_id = node.get_property(pname='reservation_info').reservation_id
                management_ip = node.management_ip

                hostname = reservation_id + '-' + name.lower()

                expected_stdout = '3D controller: NVIDIA Corporation TU104GL [Tesla T4] (rev a1)'

                #print("------------------------- Test Output ---------------------------")
                script= '#!/bin/bash  \n' \
                """lspci | grep Tesla \n"""
                #'lspci \| grep \"3D controller: NVIDIA Corporation TU104GL [Tesla T4]\" \n'
                stdout_str = self.execute_script(node_username='centos', node=node, script=script)
                #print(str(stdout_str.replace('\n','')) + " | ", end = '')
                #print("-----------------------------------------------------------------")
                #stdout_str = str(stdout.read(),'utf-8').replace('\\n','\n')
                if expected_stdout in stdout_str:
                    print('Success')
                else:
                    print('Fail')
                    #print('Fail: --{}--  --{}--'.format(expected_stdout,stdout_str))
            except Exception as e:
                print ("Error in test: Error {}".format(e))
                traceback.print_exc()



    def delete_slice(self):
        status, result = self.slice_manager.delete(slice_id=self.slice.slice_id)

        #print("Response Status {}".format(status))
        #print("Response received {}".format(result))

    def get_existing_slice(self, slice_name):
        #Get existing slice
        return_status, slices = self.slice_manager.slices(excludes=[SliceState.Dead, SliceState.Closing])

        if return_status == Status.OK:
            #for slice in slices:
            #    print("{}:".format(slice.slice_name))
            #    print("   ID         : {}".format(slice.slice_id))
            #    print("   State      : {}".format(slice.slice_state))
            #    print("   Lease End  : {}".format(slice.lease_end))
            #print("got Slice")
            pass
        else:
            print(f"Failure: {slices}")

        self.slice = list(filter(lambda x: x.slice_name == slice_name, slices))[0]
        status, self.topology = self.slice_manager.get_slice_topology(slice_object=self.slice)
        if return_status == Status.OK:
            #print("got topology")
            pass
        else:
            print(f"Failure: {self.topology}")

        #print("Slice Name : {}".format(self.slice.slice_name))
        #print("ID         : {}".format(self.slice.slice_id))
        #print("State      : {}".format(self.slice.slice_state))
        #print("Lease End  : {}".format(self.slice.lease_end))

    def run_all(self, slice_name, sites=[], create_slice=True, run_test=True, delete=True, node_count=1):
        for site, node_count in sites:
            #print("site: {}, node_count: {}".format(site,node_count))
            self.run(slice_name + "-" + site, create_slice=create_slice, run_test=run_test, delete=delete, site=site, node_count=node_count)


    def run(self, slice_name, create_slice=True, run_test=True, delete=True, site=None, node_count=1):
        """
        Run the test
        :return:
        """
        print("Tesla_T4 test, slice_name: {}, site: {}".format(slice_name,site))
        if create_slice:
            #print("Creating Slice")
            try:
                self.create_slice(slice_name,site,node_count=node_count)
                time.sleep(5)

                self.slice = self.wait_for_slice(self.slice, progress=True, timeout=300+(20*node_count))
                #time.sleep(10)

            except Exception as e:
                print("Create Slice FAILED. Error {}".format(e))
                traceback.print_exc()
            time.sleep(5*node_count)


        self.get_existing_slice(slice_name)
        #time.sleep(60)

        if run_test:
            try:
                #print("Run Test")
                self.run_test()
            except Exception as e:
                print("Run test FAILED. Error {}".format(e))
                traceback.print_exc()
        if delete:
            try:
                #print("Delete Slice")
                self.delete_slice()
            except:
                print("Delete FAILED")



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

import os
import traceback
import re

import functools
import time

import importlib.resources as pkg_resources
from typing import List

from fabrictestbed.slice_editor import Labels, ExperimentTopology, Capacities, ComponentType, ComponentModelType, ServiceType, ComponentCatalog
from fabrictestbed.slice_editor import (
    ExperimentTopology,
    Capacities
)
from fabrictestbed.slice_manager import SliceManager, Status, SliceState


from .abc_test import AbcTest

from .. import images


class ComponentTest(AbcTest):

    COMPONENT_MODELS = {  'GPU_Tesla_T4': ComponentModelType.GPU_Tesla_T4,
                          'GPU_RTX6000': ComponentModelType.GPU_RTX6000,
                          'SharedNIC_ConnectX_6': ComponentModelType.SharedNIC_ConnectX_6,
                          'SmartNIC_ConnectX_6': ComponentModelType.SmartNIC_ConnectX_6,
                          'SmartNIC_ConnectX_5': ComponentModelType.SmartNIC_ConnectX_5,
                          'NVME_P4510': ComponentModelType.NVME_P4510
                          }

    def __init__(self):
        """
        Constructor
        :return:
        """
        super().__init__()






    def create_slice(self, name_prefix='component_test', model_type=None, site_name='', node_count=1):
        image = 'default_centos_8'
        image_type = 'qcow2'
        cores = 2
        ram = 2
        disk = 2
        #model_type = ComponentModelType.GPU_Tesla_T4
        #Tesla_T4_component_type = ComponentType.Tesla_T4
        #Tesla_T4_model = 'P4510'
        #Tesla_T4_name = 'gpu1'
        component_name = 'dev1'
        component_model_name = str(model_type).split(".",1)[1]

        topology = ExperimentTopology()

        cap = Capacities(core=cores, ram=ram, disk=disk)

        slice_name = name_prefix+'-'+component_model_name+'-'+str(site_name)

        for node_num in range(0, node_count):
            #node_name=slice_name+"-"+str(node_num)
            node_name='node'+str(node_num)
            # Add node
            n1 = topology.add_node(name=node_name, site=site_name)

            # Set Properties
            n1.set_properties(capacities=cap, image_type=image_type, image_ref=image)

            # Add the PCI Tesla_T4 device
            n1.add_component(model_type=model_type, name=component_name)
            #n1.add_component(ctype=Tesla_T4_component_type, model=Tesla_T4_model, name=Tesla_T4_name)

        # Generate Slice Graph
        slice_graph = topology.serialize()

        # Request slice from Orchestrator
        return_status, slice_reservations = self.slice_manager.create(slice_name=slice_name,
                                                    slice_graph=slice_graph,
                                                    ssh_key=self.node_ssh_key)




        if return_status == Status.OK:
            slice_id = slice_reservations[0].get_slice_id()
            #print("Submitted slice creation request. Slice ID: {}".format(slice_id))
        else:
            print(f"Failure: {slice_reservations}")

        #time.sleep(30)

        return_status, slices = self.slice_manager.slices(excludes=[SliceState.Dead,SliceState.Closing])

        if return_status == Status.OK:
            self.slices.append(list(filter(lambda x: x.slice_name == slice_name, slices))[0])
            #self.slice = self.wait_for_slice(slice, progress=True)

        #print()
        #print("Slice Name : {}".format(self.slice.slice_name))
        #print("ID         : {}".format(self.slice.slice_id))
        #print("State      : {}".format(self.slice.slice_state))
        #print("Lease End  : {}".format(self.slice.lease_end))

    def config_test_GPU_RTX6000(self, node):

        node_name = node.name
        try:
            name = node_name
            management_ip = node.management_ip

            #print("------------------------- Test Output ---------------------------")
            script= '#!/bin/bash  \n' \
            'sudo yum install -y pciutils \n'
            try:
                stdout_str = self.execute_script(node_username='centos', node=node, script=script)
                print("Config Done | ", end = '')
            except Exception as e:
                print("Config Fail {} | ".format(str(e)), end = '')
        except Exception as e:
            print ("Error in test: Error {}".format(e), end = '')
            traceback.print_exc()

    def run_test_GPU_RTX6000(self, slice):
        return_status, topology = self.slice_manager.get_slice_topology(slice_object=slice)
        if return_status != Status.OK:
            raise Exception("config_test failed to get topology. slice; {}, error {}".format(str(slice),str(topology)))

        for node_name, node in topology.nodes.items():
            print(str(slice.slice_name) + " | " + str(node_name) + " | ", end = '')
            self.config_test_GPU_RTX6000(node)
            try:
                print("{}".format(str(node.get_property(pname='label_allocations').instance_parent)).replace('\n','') + " | ", end = '')
                print("{}".format(str(node.management_ip)).replace('\n','') + " | ", end = '')

                name = node_name
                reservation_id = node.get_property(pname='reservation_info').reservation_id
                management_ip = node.management_ip

                hostname = reservation_id + '-' + name.lower()

                expected_stdout = '3D controller: NVIDIA Corporation TU102GL [Quadro RTX 6000/8000] (rev a1)'

                script= '#!/bin/bash  \n' \
                """lspci | grep NVIDIA \n"""
                #'lspci \| grep \"3D controller: NVIDIA Corporation TU104GL [Tesla T4]\" \n'
                stdout_str = self.execute_script(node_username='centos', node=node, script=script)
                print(str(stdout_str.replace('\n',''))[0:45] + "... | ", end = '')
                if expected_stdout in stdout_str:
                    print('Success')
                else:
                    print('Fail')
                    print(stdout_str)
            except Exception as e:
                print ("Error in test: Error {}".format(e))
                traceback.print_exc()


    def config_test_NVME_P4510(self, node):

        node_name = node.name
        try:
            name = node_name
            management_ip = node.management_ip

            #print("------------------------- Test Output ---------------------------")
            script= '#!/bin/bash  \n' \
            'sudo yum install -y pciutils \n'
            try:
                stdout_str = self.execute_script(node_username='centos', node=node, script=script)
                print("Config Done | ", end = '')
            except Exception as e:
                print("Config Fail {} | ".format(str(e)), end = '')
        except Exception as e:
            print ("Error in test: Error {}".format(e), end = '')
            traceback.print_exc()

    def run_test_NVME_P4510(self, slice):
        return_status, topology = self.slice_manager.get_slice_topology(slice_object=slice)
        if return_status != Status.OK:
            raise Exception("config_test failed to get topology. slice; {}, error {}".format(str(slice),str(topology)))

        for node_name, node in topology.nodes.items():
            print(str(slice.slice_name) + " | " + str(node_name) + " | ", end = '')
            self.config_test_NVME_P4510(node)
            try:
                print("{}".format(str(node.get_property(pname='label_allocations').instance_parent)).replace('\n','') + " | ", end = '')
                print("{}".format(str(node.management_ip)).replace('\n','') + " | ", end = '')

                name = node_name
                reservation_id = node.get_property(pname='reservation_info').reservation_id
                management_ip = node.management_ip

                hostname = reservation_id + '-' + name.lower()

                #There are two types of NVMe
                expected_stdout1 = 'Non-Volatile memory controller: Intel Corporation NVMe Datacenter SSD [3DNAND, Beta Rock Controller]'
                expected_stdout2 = 'Non-Volatile memory controller: Toshiba Corporation NVMe SSD Controller Cx5 (rev 01)'
                script= '#!/bin/bash  \n' \
                """lspci | grep NVMe \n"""
                #'lspci \| grep \"3D controller: NVIDIA Corporation TU104GL [Tesla T4]\" \n'
                stdout_str = self.execute_script(node_username='centos', node=node, script=script)
                print(str(stdout_str.replace('\n',''))[0:45] + "... | ", end = '')
                if expected_stdout1 in stdout_str or expected_stdout2 in stdout_str :
                    print('Success')
                else:
                    print('Fail')
                    print(stdout_str)
            except Exception as e:
                print ("Error in test: Error {}".format(e))
                traceback.print_exc()


    def config_test_SmartNIC_ConnectX_5(self, node):

        node_name = node.name
        try:
            name = node_name
            management_ip = node.management_ip

            #print("------------------------- Test Output ---------------------------")
            script= '#!/bin/bash  \n' \
            'sudo yum install -y pciutils \n'
            try:
                stdout_str = self.execute_script(node_username='centos', node=node, script=script)
                print("Config Done | ", end = '')
            except Exception as e:
                print("Config Fail {} | ".format(str(e)), end = '')
        except Exception as e:
            print ("Error in test: Error {}".format(e), end = '')
            traceback.print_exc()

    def run_test_SmartNIC_ConnectX_5(self, slice):
        return_status, topology = self.slice_manager.get_slice_topology(slice_object=slice)
        if return_status != Status.OK:
            raise Exception("config_test failed to get topology. slice; {}, error {}".format(str(slice),str(topology)))

        for node_name, node in topology.nodes.items():
            print(str(slice.slice_name) + " | " + str(node_name) + " | ", end = '')
            self.config_test_SmartNIC_ConnectX_5(node)
            try:
                print("{}".format(str(node.get_property(pname='label_allocations').instance_parent)).replace('\n','') + " | ", end = '')
                print("{}".format(str(node.management_ip)).replace('\n','') + " | ", end = '')


                name = node_name
                reservation_id = node.get_property(pname='reservation_info').reservation_id
                management_ip = node.management_ip

                hostname = reservation_id + '-' + name.lower()

                expected_stdout = 'Ethernet controller: Mellanox Technologies MT27800 Family [ConnectX-5]'

                script= '#!/bin/bash  \n' \
                """lspci | grep ConnectX \n"""
                #'lspci \| grep \"3D controller: NVIDIA Corporation TU104GL [Tesla T4]\" \n'
                stdout_str = self.execute_script(node_username='centos', node=node, script=script)
                print(str(stdout_str.replace('\n',''))[0:45] + "... | ", end = '')
                if expected_stdout in stdout_str:
                    print('Success')
                else:
                    print('Fail')
                    print(stdout_str)
            except Exception as e:
                print ("Error in test: Error {}".format(e))
                traceback.print_exc()

    def config_test_SmartNIC_ConnectX_6(self, node):

        node_name = node.name
        try:
            name = node_name
            management_ip = node.management_ip

            #print("------------------------- Test Output ---------------------------")
            script= '#!/bin/bash  \n' \
            'sudo yum install -y pciutils \n'
            try:
                stdout_str = self.execute_script(node_username='centos', node=node, script=script)
                print("Config Done | ", end = '')
            except Exception as e:
                print("Config Fail {} | ".format(str(e)), end = '')
        except Exception as e:
            print ("Error in test: Error {}".format(e), end = '')
            traceback.print_exc()

    def run_test_SmartNIC_ConnectX_6(self, slice):
        return_status, topology = self.slice_manager.get_slice_topology(slice_object=slice)
        if return_status != Status.OK:
            raise Exception("config_test failed to get topology. slice; {}, error {}".format(str(slice),str(topology)))

        for node_name, node in topology.nodes.items():
            #print(" Test: SmartNIC_ConnectX_6 | " + str(slice.slice_name) + " | " + str(node_name) + " | ", end = '')
            print(str(slice.slice_name) + " | " + str(node_name) + " | ", end = '')

            self.config_test_SmartNIC_ConnectX_6(node)
            try:
                print("{}".format(str(node.get_property(pname='label_allocations').instance_parent)).replace('\n','') + " | ", end = '')
                print("{}".format(str(node.management_ip)).replace('\n','') + " | ", end = '')

                name = node_name
                reservation_id = node.get_property(pname='reservation_info').reservation_id
                management_ip = node.management_ip

                hostname = reservation_id + '-' + name.lower()

                expected_stdout = 'Ethernet controller: Mellanox Technologies MT28908 Family [ConnectX-6]'

                script= '#!/bin/bash  \n' \
                """lspci | grep ConnectX-6 \n"""
                #'lspci \| grep \"3D controller: NVIDIA Corporation TU104GL [Tesla T4]\" \n'
                stdout_str = self.execute_script(node_username='centos', node=node, script=script)
                print(str(stdout_str.replace('\n',''))[0:45] + "...  | ", end = '')
                if expected_stdout in stdout_str:
                    print('Success')
                else:
                    print('Fail')
                    print(stdout_str)
            except Exception as e:
                print ("Error in test: Error {}".format(e))
                traceback.print_exc()


    def config_test_SharedNIC_ConnectX_6(self, node):

        node_name = node.name
        try:
            name = node_name
            management_ip = node.management_ip

            #print("------------------------- Test Output ---------------------------")
            script= '#!/bin/bash  \n' \
            'sudo yum install -y pciutils \n'
            try:
                stdout_str = self.execute_script(node_username='centos', node=node, script=script)
                print("Config Done | ", end = '')
            except Exception as e:
                print("Config Fail {} | ".format(str(e)), end = '')
        except Exception as e:
            print ("Error in test: Error {}".format(e), end = '')
            traceback.print_exc()

    def run_test_SharedNIC_ConnectX_6(self, slice):
        return_status, topology = self.slice_manager.get_slice_topology(slice_object=slice)
        if return_status != Status.OK:
            raise Exception("config_test failed to get topology. slice; {}, error {}".format(str(slice),str(topology)))

        for node_name, node in topology.nodes.items():
            print(str(slice.slice_name) + " | " + str(node_name) + " | ", end = '')
            self.config_test_SharedNIC_ConnectX_6(node)
            try:
                print("{}".format(str(node.get_property(pname='label_allocations').instance_parent)).replace('\n','') + " | ", end = '')
                print("{}".format(str(node.management_ip)).replace('\n','') + " | ", end = '')

                name = node_name
                reservation_id = node.get_property(pname='reservation_info').reservation_id
                management_ip = node.management_ip

                hostname = reservation_id + '-' + name.lower()

                expected_stdout = 'Ethernet controller: Mellanox Technologies MT28908 Family [ConnectX-6 Virtual Function]'

                script= '#!/bin/bash  \n' \
                """lspci | grep ConnectX-6 \n"""
                #'lspci \| grep \"3D controller: NVIDIA Corporation TU104GL [Tesla T4]\" \n'
                stdout_str = self.execute_script(node_username='centos', node=node, script=script)
                print(str(stdout_str.replace('\n',''))[0:45] + "... | ", end = '')
                if expected_stdout in stdout_str:
                    print('Success')
                else:
                    print('Fail')
                    print(stdout_str)
            except Exception as e:
                print ("Error in test: Error {}".format(e))
                traceback.print_exc()


    def config_test_GPU_Tesla_T4(self, node):

        node_name = node.name
        #print("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
        #print("Node:")
        #print("Config | ", end = '')
        #print(str(node_name) + " | ", end = '')
        try:
            #print("{}".format(str(node_name)) + " | ", end = '')

            #print("   Name              : {}".format(node_name))
            #print("   Cores             : {}".format(node.get_property(pname='capacity_allocations').core))
            #print("   RAM               : {}".format(node.get_property(pname='capacity_allocations').ram))
            #print("   Disk              : {}".format(node.get_property(pname='capacity_allocations').disk))
            #print("   Image             : {}".format(node.image_ref))
            #print("   Image Type        : {}".format(node.image_type))
            #print("   Host              : {}".format(node.get_property(pname='label_allocations').instance_parent))
            #print("{}".format(str(node.get_property(pname='label_allocations').instance_parent)).replace('\n','') + " | ", end = '')
            #print("   Site              : {}".format(node.site))
            #print("{}".format(str(node.management_ip)).replace('\n','') + " | ", end = '')

            #print("   Management IP     : {}".format(node.management_ip))
            #print("   Reservation ID    : {}".format(node.get_property(pname='reservation_info').reservation_id))
            #print("   Reservation State : {}".format(node.get_property(pname='reservation_info').reservation_state))
            #print("   Components        : {}".format(node.components))
            #print("   Interfaces        : {}".format(node.interfaces))
            #print()
            name = node_name
            management_ip = node.management_ip

            #print("------------------------- Test Output ---------------------------")
            script= '#!/bin/bash  \n' \
            'sudo yum install -y pciutils \n'
            try:
                stdout_str = self.execute_script(node_username='centos', node=node, script=script)
                print("Config Done | ", end = '')
            except Exception as e:
                print("Config Fail {} | ".format(str(e)), end = '')
            #print(str(stdout_str.replace('\n','')) + " | ", end = '')
            #print("-----------------------------------------------------------------")
            #stdout_str = str(stdout.read(),'utf-8').replace('\\n','\n')
        except Exception as e:
            print ("Error in test: Error {}".format(e), end = '')
            traceback.print_exc()




    def run_test_GPU_Tesla_T4(self, slice):
        #self.config_test(slice)

        return_status, topology = self.slice_manager.get_slice_topology(slice_object=slice)
        if return_status != Status.OK:
            raise Exception("config_test failed to get topology. slice; {}, error {}".format(str(slice),str(topology)))


        for node_name, node in topology.nodes.items():
            #print("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX")
            #print("Node:")
            #print(str(node_name) + " | ", end = '')
            print(str(slice.slice_name) + " | " + str(node_name) + " | ", end = '')
            self.config_test_GPU_Tesla_T4(node)
            try:
                #print("   Name              : {}".format(node_name))
                #print("   Cores             : {}".format(node.get_property(pname='capacity_allocations').core))
                #print("   RAM               : {}".format(node.get_property(pname='capacity_allocations').ram))
                #print("   Disk              : {}".format(node.get_property(pname='capacity_allocations').disk))
                #print("   Image             : {}".format(node.image_ref))
                #print("   Image Type        : {}".format(node.image_type))
                #print("   Host              : {}".format(node.get_property(pname='label_allocations').instance_parent))
                #print("   Site              : {}".format(node.site))
                #print("   Management IP     : {}".format(node.management_ip))
                #print("   Reservation ID    : {}".format(node.get_property(pname='reservation_info').reservation_id))
                #print("   Reservation State : {}".format(node.get_property(pname='reservation_info').reservation_state))
                #print("   Components        : {}".format(node.components))
                #print("   Interfaces        : {}".format(node.interfaces))
                #print()
                print("{}".format(str(node.get_property(pname='label_allocations').instance_parent)).replace('\n','') + " | ", end = '')
                print("{}".format(str(node.management_ip)).replace('\n','') + " | ", end = '')


                name = node_name
                reservation_id = node.get_property(pname='reservation_info').reservation_id
                management_ip = node.management_ip

                hostname = reservation_id + '-' + name.lower()

                expected_stdout = '3D controller: NVIDIA Corporation TU104GL [Tesla T4] (rev a1)'

                #print("------------------------- Test Output ---------------------------")
                script= '#!/bin/bash  \n' \
                """lspci | grep Tesla \n"""
                #'lspci \| grep \"3D controller: NVIDIA Corporation TU104GL [Tesla T4]\" \n'
                stdout_str = self.execute_script(node_username='centos', node=node, script=script)
                print(str(stdout_str.replace('\n',''))[0:45] + "... | ", end = '')
                #print("-----------------------------------------------------------------")
                #stdout_str = str(stdout.read(),'utf-8').replace('\\n','\n')
                if expected_stdout in stdout_str:
                    print('Success')
                else:
                    print('Fail')
                    print(stdout_str)
            except Exception as e:
                print ("Error in test: Error {}".format(e))
                traceback.print_exc()



    def delete_slice(self):
        status, result = self.slice_manager.delete(slice_id=self.slice.slice_id)

        #print("Response Status {}".format(status))
        #print("Response received {}".format(result))

    def get_existing_slice(self, slice_name):
        #Get existing slice
        return_status, slices = self.slice_manager.slices(excludes=[SliceState.Dead, SliceState.Closing])

        if return_status == Status.OK:
            #for slice in slices:
            #    print("{}:".format(slice.slice_name))
            #    print("   ID         : {}".format(slice.slice_id))
            #    print("   State      : {}".format(slice.slice_state))
            #    print("   Lease End  : {}".format(slice.lease_end))
            #print("got Slice")
            pass
        else:
            print(f"Failure: {slices}")

        self.slice = list(filter(lambda x: x.slice_name == slice_name, slices))[0]
        status, self.topology = self.slice_manager.get_slice_topology(slice_object=self.slice)
        if return_status == Status.OK:
            #print("got topology")
            pass
        else:
            print(f"Failure: {self.topology}")

        #print("Slice Name : {}".format(self.slice.slice_name))
        #print("ID         : {}".format(self.slice.slice_id))
        #print("State      : {}".format(self.slice.slice_state))
        #print("Lease End  : {}".format(self.slice.lease_end))




    def run_all(self, test_name='component_test', tests=[], create_slice=True, run_test=True, delete=True, node_count=1):
        self.test_name = test_name

        # Create slices
        if create_slice:
            for site, model_type, node_count in tests:
                print("Create slice:  site: {}, component: {}, node_count: {}. Nodes: ".format(site,model_type,node_count), end='')
                #print("Creating Slice")
                for node_num in range(0,node_count):
                    try:
                        print(" {}".format(node_num),end='')

                        self.create_slice(name_prefix=test_name+"-"+str(node_num),model_type=ComponentTest.COMPONENT_MODELS[model_type],site_name=site,node_count=1)
                        #time.sleep(2)

                        #self.slice = self.wait_for_slice(self.slice, progress=True, timeout=300+(20*node_count))
                        #time.sleep(10)

                    except Exception as e:
                        print("Create Slice FAILED. Error {}".format(e))
                        traceback.print_exc()
                print(', Done')

            for slice in self.slices:
                try:
                    s = self.wait_for_slice(slice, progress=True, timeout=300)
                except Exception as e:
                    print("Slice failed wait_for_slice. {}".format(slice))
                    print(str(e))
                    #self.slices.remove(slice)


            time.sleep(120)
            #test ssh conectivity
            for slice in self.slices:
                try:
                    self.run_ssh_test(slice)
                except Exception as e:
                    #print("Failed to get topology: {}".format(slice))
                    #print(str(e))
                    print(" Test: ssh | " + str(slice.slice_name) + " | slice closing | Fail ")


        #Run tests
        if run_test:
            print("Run tests")
            for site, model_type, node_count in tests:
                #print("Run test:  site: {}, component: {}, node_count: {}".format(site,model_type,node_count))
                #print("Creating Slice")
                for node_num in range(0,node_count):
                    try:
                        #print("Running test  site: {}, component: {}, node_num: {}".format(site,model_type,node_num))
                        name_prefix=test_name+"-"+str(node_num)
                        component_model_name = str(model_type) #.split(".",1)[1]
                        slice_name = name_prefix+'-'+component_model_name+'-'+str(site)
                        #print("Running test, slice_name: {}".format(slice_name))
                        from fabrictestbed_extensions.utils.slice import SliceUtils

                        slice = SliceUtils.get_slice(slice_name=slice_name, slice_manager=self.slice_manager)

                        if model_type == 'GPU_Tesla_T4':
                            self.run_test_GPU_Tesla_T4(slice)
                        elif model_type == 'GPU_RTX6000':
                            self.run_test_GPU_RTX6000(slice)
                        elif model_type == 'SharedNIC_ConnectX_6':
                            self.run_test_SharedNIC_ConnectX_6(slice)
                        elif model_type == 'SmartNIC_ConnectX_6':
                            self.run_test_SmartNIC_ConnectX_6(slice)
                        elif model_type == 'SmartNIC_ConnectX_5':
                            self.run_test_SmartNIC_ConnectX_5(slice)
                        elif model_type == 'NVME_P4510':
                            self.run_test_NVME_P4510(slice)
                        else:
                            print(" Test: XXX | " + str(slice_name) + " | unknown test | Skipped")

                    except Exception as e:
                        print(" Test: XXX | " + str(slice_name) + " | " + str(e) + " | Failed")

                        #print("Run test FAILED. Error {}".format(e))
                        #traceback.print_exc()







        #self.run(slice_name + "-" + site, create_slice=False, run_test=run_test, delete=delete, site=site, node_count=node_count)


    def run(self, test_name='component_test', create_slice=True, run_test=True, delete=True, site=None, node_count=1):
        """
        Run the test
        :return:
        """
        self.test_name = test_name
        print("Tesla_T4 test, slice_name: {}, site: {}".format(slice_name,site))
        if create_slice:
            #print("Creating Slice")
            try:
                self.create_slice(model_type=ComponentModelType.GPU_Tesla_T4,site_name=site,node_count=node_count)
                #time.sleep(5)

                self.slice = self.wait_for_slice(self.slice, progress=True, timeout=300+(20*node_count))
                #time.sleep(10)

            except Exception as e:
                print("Create Slice FAILED. Error {}".format(e))
                traceback.print_exc()
            time.sleep(5*node_count)


        self.get_existing_slice(slice_name)
        #time.sleep(60)

        if run_test:
            try:
                #print("Run Test")
                self.run_test_tesla_t4()
            except Exception as e:
                print("Run test FAILED. Error {}".format(e))
                traceback.print_exc()
        if delete:
            try:
                #print("Delete Slice")
                self.delete_slice()
            except:
                print("Delete FAILED")



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)

from tabulate import tabulate
from typing import List

from fabrictestbed.slice_editor import ComponentModelType, Component

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from fabrictestbed_extensions.fablib.slice import Slice
    from fabrictestbed_extensions.fablib.node import Node
    from fabrictestbed_extensions.fablib.interface import Interface
    from fabrictestbed.slice_editor import Component as FimComponent


class Component:
    component_model_map = { 'NIC_Basic': ComponentModelType.SharedNIC_ConnectX_6,
                            'NIC_ConnectX_6': ComponentModelType.SmartNIC_ConnectX_6,
                            'NIC_ConnectX_5': ComponentModelType.SmartNIC_ConnectX_5,
                            'NVME_P4510': ComponentModelType.NVME_P4510,
                            'GPU_TeslaT4': ComponentModelType.GPU_Tesla_T4,
                            'GPU_RTX6000': ComponentModelType.GPU_RTX6000
                            }

    def __str__(self):
        """
        Creates a tabulated string describing the properties of the component.

        Intended for printing component information.

        :return: Tabulated string of component information
        :rtype: String
        """
        table = [   [ "Name", self.get_name() ],
                    [ "Details", self.get_details() ],
                    [ "Disk (G)", self.get_disk() ],
                    [ "Units", self.get_unit() ],
                    [ "PCI Address", self.get_pci_addr() ],
                    [ "Model", self.get_model() ],
                    [ "Type", self.get_type() ],
                    ]

        return tabulate(table)

    def list_interfaces(self) -> List[str]:
        """
        Creates a tabulated string describing all components in the slice.

        Intended for printing a list of all components.

        :return: Tabulated srting of all components information
        :rtype: String
        """
        table = []
        for iface in self.get_interfaces():
            network_name = ""
            if iface.get_network():
                network_name = iface.get_network().get_name()

            table.append( [     iface.get_name(),
                                network_name,
                                iface.get_bandwidth(),
                                iface.get_vlan(),
                                iface.get_mac(),
                                iface.get_physical_os_interface_name(),
                                iface.get_os_interface(),
                                ] )

        return tabulate(table, headers=["Name", "Network", "Bandwidth", "VLAN", "MAC", "Physical OS Interface", "OS Interface" ])

    @staticmethod
    def calculate_name(node: Node = None, name: str = None) -> str:
        """
        Not intended for API use
        """
        # Hack to make it possile to find interfaces
        return f"{node.get_name()}-{name}"

    @staticmethod
    def new_component(node: Node = None, model: str = None, name: str = None):
        """
        Not intended for API use

        Creates a new FIM component on the fablib node inputted.

        :param node: the fablib node to build the component on
        :type node: Node
        :param model: the name of the component type to build
        :type model: str
        :param name: the name of the new component
        :type name: str
        :return: the new fablib compoent
        :rtype: Component
        """
        # Hack to make it possile to find interfaces
        name = Component.calculate_name(node=node, name=name)

        return Component(node=node, fim_component=node.fim_node.add_component(
            model_type=Component.component_model_map[model], name=name))
        # return Component(node = node, model=model, name=name)

    def __init__(self, node: Node = None, fim_component: FimComponent = None):
        """
        Not intended for API use

        Constructor. Sets the FIM component and fablib node to the inputted values.

        :param node: the fablib node to build the component on
        :type node: Node
        :param fim_component: the FIM component this object represents
        :type fim_component: FIMComponent
        """
        super().__init__()
        self.fim_component = fim_component
        self.node = node

    def get_interfaces(self) -> List[Interface]:
        """
        Gets the interfaces attached to this fablib component's FABRIC component.

        :return: a list of the interfaces on this component.
        :rtype: List[Interface]
        """

        from fabrictestbed_extensions.fablib.interface import Interface
        ifaces = []
        for fim_interface in self.get_fim_component().interface_list:
            ifaces.append(Interface(component=self, fim_interface=fim_interface))

        return ifaces

    def get_fim_component(self) -> FimComponent:
        """
        Not intended for API use

        Gets the FABRIC component this fablib component represents.

        :return: the FABRIC component on this component
        :rtype: FIMComponent
        """
        return self.fim_component

    def get_slice(self) -> Slice:
        """
        Gets the fablib slice associated with this component's node.

        :return: the slice this component is on
        :rtype: Slice
        """
        return self.node.get_slice()

    def get_node(self) -> Node:
        """
        Gets the fablib node this component is associated with.

        :return: the node this component is on
        :rtype: Node
        """
        return self.node

    def get_site(self) -> str:
        """
        Gets the name of the site this component's node is on.

        :return: the site name this node is on
        :rtype: String
        """
        return self.node.get_site()

    def get_name(self) -> str:
        """
        Gets the name of this component from the FABRIC component.

        :return: the name of this component
        :rtype: str
        """
        return self.get_fim_component().name

    def get_details(self) -> str:
        """
        Not intended for API use
        """
        return self.get_fim_component().details

    def get_disk(self) -> int:
        """
        Gets the amount of disk space on this component.

        :return: this component's disk space
        :rtype: int
        """
        return self.get_fim_component().get_property(pname='capacity_allocations').disk

    def get_unit(self) -> int:
        """
        Get unit count for this component.

        :return: unit
        :rtype: int
        """
        return self.get_fim_component().get_property(pname='capacity_allocations').unit

    def get_pci_addr(self) -> str:
        """
        Get the PIC device ID for this component.

        :return: PCI device ID
        :rtype: String
        """
        return self.get_fim_component().get_property(pname='label_allocations').bdf

    def get_model(self) -> str:
        """
        Get FABlib model name for this component.

        :return: FABlib model name
        :rtype: String
        """
        #TODO: This a hack that need a real fix
        if str(self.get_type()) == "SmartNIC" and str(self.get_fim_model()) == "ConnectX-6":
            return 'NIC_ConnectX_6'
        elif str(self.get_type()) == "SmartNIC" and str(self.get_fim_model()) == "ConnectX-5":
            return 'NIC_ConnectX_5'
        elif str(self.get_type()) == "NVME"  and str(self.get_fim_model()) == "P4510":
            return 'NVME_P4510'
        elif str(self.get_type())== "GPU"  and str(self.get_fim_model()) == "Tesla T4":
            return 'GPU_TeslaT4'
        elif str(self.get_type()) == "GPU"  and str(self.get_fim_model()) == "RTX6000":
            return 'GPU_RTX6000'
        elif str(self.get_type()) == "SharedNIC"  and str(self.get_fim_model()) == "ConnectX-6":
            return 'NIC_Basic'
        else:
            return None

    def get_reservation_id(self) -> str or None:
        """
        Get reservation ID for this component.

        :return:  reservation ID
        :rtype: String
        """
        try:
            #This does not work
            #print(f"{self.get_fim_component()}")
            return self.get_fim_component().get_property(pname='reservation_info').reservation_id
        except:
            return None

    def get_reservation_state(self) -> str or None:
        """
        Get reservation state for this component.

        :return:  reservation state
        :rtype: String
        """
        try:
            return self.get_fim_component().get_property(pname='reservation_info').reservation_state
        except:
            return None

    def get_error_message(self) -> str:
        """
        Get error message for this component.

        :return:  reservation state
        :rtype: String
        """
        try:
            return self.get_fim_component().get_property(pname='reservation_info').error_message
        except:
            return ""

    def get_fim_model(self) -> str:
        """
        Not for API use
        """
        return self.get_fim_component().model

    def get_type(self) -> str:
        """
        Not for API use

        Gets the type of this component.

        :return: the type of component
        :rtype: str
        """
        return self.get_fim_component().type

    def configure_nvme(self, mount_point='/mnt/nvme_mount'):
        """
        Configure the NVMe drive.

        Note this works but may be reorganzied. 

        :param mount_point: The mount point in the filesystem. Default = /mnt/nvme_mount
        :type mount_point: String
        """
        output = []
        try:
            output.append(self.node.execute('sudo fdisk -l /dev/nvme*'))
            output.append(self.node.execute('sudo parted -s /dev/nvme0n1 mklabel gpt'))
            output.append(self.node.execute('sudo parted -s /dev/nvme0n1 print'))
            output.append(self.node.execute('sudo parted -s /dev/nvme0n1 print unit MB print free'))
            output.append(self.node.execute('sudo parted -s --align optimal /dev/nvme0n1 mkpart primary ext4 0% 960197MB'))
            output.append(self.node.execute('lsblk /dev/nvme0n1'))
            output.append(self.node.execute('sudo mkfs.ext4 /dev/nvme0n1p1'))
            output.append(self.node.execute(f'sudo mkdir {mount_point} && sudo mount /dev/nvme0n1p1 {mount_point}'))
            output.append(self.node.execute(f'df -h {mount_point}'))
        except Exception as e:
            print(f"config_nvme Fail: {self.get_name()}")
            #traceback.print_exc()
            raise Exception(str(output))

        return output



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author: Paul Ruth (pruth@renci.org)
import json
import time
import paramiko
import logging

from tabulate import tabulate

from typing import List, Union

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from fabrictestbed_extensions.fablib.slice import Slice
    from fabric_cf.orchestrator.swagger_client import Sliver as OrchestratorSliver

from fim.slivers.network_service import NSLayer

from fabrictestbed.slice_editor import Labels, CapacityHints, ServiceType
from fabrictestbed.slice_editor import Capacities
from ipaddress import ip_address, IPv4Address, IPv6Address, IPv4Network, IPv6Network

from fabrictestbed_extensions.fablib.component import Component
from fabrictestbed_extensions.fablib.fablib import fablib
from fabrictestbed_extensions.fablib.interface import Interface
from fabrictestbed.slice_editor import Node as FimNode



class Node:
    default_cores = 2
    default_ram = 8
    default_disk = 10
    default_image = 'default_rocky_8'

    def __init__(self, slice: Slice, node: FimNode):
        """
        Constructor. Sets the fablib slice and FIM node based on arguments.
        :param slice: the fablib slice to have this node on
        :type slice: Slice
        :param node: the FIM node that this Node represents
        :type node: Node
        """
        super().__init__()
        self.fim_node = node
        self.slice = slice
        self.host = None

        # Try to set the username.
        try:
            self.set_username()
        except:
            self.username = None

        try:
            self.sliver = slice.get_sliver(reservation_id=self.get_reservation_id())
        except:
            self.sliver = None

        logging.getLogger("paramiko").setLevel(logging.WARNING)

    def __str__(self):
        """
        Creates a tabulated string describing the properties of the node.
        Intended for printing node information.
        :return: Tabulated string of node information
        :rtype: String
        """
        table = [ ["ID", self.get_reservation_id()],
            ["Name", self.get_name()],
            ["Cores", self.get_cores()],
            ["RAM", self.get_ram()],
            ["Disk", self.get_disk()],
            ["Image", self.get_image()],
            ["Image Type", self.get_image_type()],
            ["Host", self.get_host()],
            ["Site", self.get_site()],
            ["Management IP", self.get_management_ip()],
            ["Reservation State", self.get_reservation_state()],
            ["Error Message", self.get_error_message()],
            ["SSH Command ", self.get_ssh_command()],
            ]

        return tabulate(table) #, headers=["Property", "Value"])

    def get_sliver(self) -> OrchestratorSliver:
        """
        Not intended as API call
        Gets the node SM sliver
        :return: SM sliver for the node
        :rtype: Sliver
        """
        return self.sliver

    @staticmethod
    def new_node(slice: Slice = None, name: str = None, site: str = None, avoid: List[str] = []):
        """
        Not intended for API call. See: Slice.add_node()
        Creates a new FABRIC node and returns a fablib node with the new node.
        :param slice: the fablib slice to build the new node on
        :type slice: Slice
        :param name: the name of the new node
        :type name: str
        :param site: the name of the site to build the node on
        :type site: str
        :param avoid: a list of node names to avoid
        :type avoid: List[str]
        :return: a new fablib node
        :rtype: Node
        """
        if site is None:
            [site] = fablib.get_random_sites(avoid=avoid)

        logging.info(f"Adding node: {name}, slice: {slice.get_name()}, site: {site}")
        node = Node(slice, slice.topology.add_node(name=name, site=site))
        node.set_capacities(cores=Node.default_cores, ram=Node.default_ram, disk=Node.default_disk)
        node.set_image(Node.default_image)

        return node

    @staticmethod
    def get_node(slice: Slice = None, node=None):
        """
        Not intended for API call.
        Returns a new fablib node using existing FABRIC resources.
        :param slice: the fablib slice storing the existing node
        :type slice: Slice
        :param node: the FIM node stored in this fablib node
        :type node: Node
        :return: a new fablib node storing resources
        :rtype: Node
        """
        return Node(slice, node)

    def get_fim_node(self) -> FimNode:
        """
        Not intended for API call.
        Gets the FABRIC node associated with this fablib node.
        :return: the real FABRIC node
        :rtype: FIMNode
        """
        return self.fim_node

    def set_capacities(self, cores: int = 2, ram: int = 2, disk: int = 10):
        """
        Sets the capacities of the FABRIC node.
        :param cores: the number of cores to set on this node
        :type cores: int
        :param ram: the amount of RAM to set on this node
        :type ram: int
        :param disk: the amount of disk space to set on this node
        :type disk: int
        """
        cores = int(cores)
        ram = int(ram)
        disk = int(disk)

        cap = Capacities(core=cores, ram=ram, disk=disk)
        self.get_fim_node().set_properties(capacities=cap)

    def set_instance_type(self, instance_type: str):
        """
        Sets the instance type of this fablib node on the FABRIC node.
        :param instance_type: the name of the instance type to set
        :type instance_type: String
        """
        self.get_fim_node().set_properties(capacity_hints=CapacityHints(instance_type=instance_type))

    def set_username(self, username: str = None):
        """
        Not intended as an API call.
        Sets this fablib node's username
        Optional username parameter. The username likely should be picked
        to match the image type.
        :param instance_type: (optional) the name of the instance type to set
        :type instance_type: String
        """
        if username != None:
            self.username = username
        elif 'centos' in self.get_image():
            self.username = 'centos'
        elif 'ubuntu' in self.get_image():
            self.username = 'ubuntu'
        elif 'rocky'in self.get_image():
            self.username = 'rocky'
        elif 'fedora'in self.get_image():
            self.username = 'fedora'
        elif 'cirros'in self.get_image():
            self.username = 'cirros'
        elif 'debian'in self.get_image():
            self.username = 'debian'
        elif 'freebsd'in self.get_image():
            self.username = 'freebsd'
        elif 'openbsd'in self.get_image():
            self.username = 'openbsd'
        else:
            self.username = None

    def set_image(self, image: str, username: str  = None, image_type: str = 'qcow2'):
        """
        Sets the image information of this fablib node on the FABRIC node.
        :param image: the image reference to set
        :type image: String
        :param username: the username of this fablib node. Currently unused.
        :type username: String
        :param image_type: the image type to set
        :type image_type: String
        """
        self.get_fim_node().set_properties(image_type=image_type, image_ref=image)
        self.set_username(username=username)

    def set_host(self, host_name: str = None):
        """
        Sets the hostname of this fablib node on the FABRIC node.
        :param host_name: the hostname. example: host_name='renc-w2.fabric-testbed.net'
        :type host_name: String
        """
        # example: host_name='renc-w2.fabric-testbed.net'
        labels = Labels()
        labels.instance_parent = host_name
        self.get_fim_node().set_properties(labels=labels)

        # set an attribute used to get host before Submit
        self.host = host_name

    def get_slice(self) -> Slice:
        """
        Gets the fablib slice associated with this node.
        :return: the fablib slice on this node
        :rtype: Slice
        """
        return self.slice

    def get_name(self) -> str or None:
        """
        Gets the name of the FABRIC node.
        :return: the name of the node
        :rtype: String
        """
        try:
            return self.get_fim_node().name
        except:
            return None

    def get_cores(self) -> int or None:
        """
        Gets the number of cores on the FABRIC node.
        :return: the number of cores on the node
        :rtype: int
        """
        try:
            return self.get_fim_node().get_property(pname='capacity_allocations').core
        except:
            return None

    def get_ram(self) -> int or None:
        """
        Gets the amount of RAM on the FABRIC node.
        :return: the amount of RAM on the node
        :rtype: int
        """
        try:
            return self.get_fim_node().get_property(pname='capacity_allocations').ram
        except:
            return None

    def get_disk(self) -> int or None:
        """
        Gets the amount of disk space on the FABRIC node.
        :return: the amount of disk space on the node
        :rtype: int
        """
        try:
            return self.get_fim_node().get_property(pname='capacity_allocations').disk
        except:
            return None

    def get_image(self) -> str or None:
        """
        Gets the image reference on the FABRIC node.
        :return: the image reference on the node
        :rtype: String
        """
        try:
            return self.get_fim_node().image_ref
        except:
            return None

    def get_image_type(self) -> str or None:
        """
        Gets the image type on the FABRIC node.
        :return: the image type on the node
        :rtype: String
        """
        try:
            return self.get_fim_node().image_type
        except:
            return None

    def get_host(self) -> str or None:
        """
        Gets the hostname on the FABRIC node.
        :return: the hostname on the node
        :rtype: String
        """
        try:
            try:
                #If we set the host but have not yet submitted
                return self.host
            except:
                pass
            return self.get_fim_node().get_property(pname='label_allocations').instance_parent
        except:
            return None

    def get_site(self) -> str or None:
        """
        Gets the sitename on the FABRIC node.
        :return: the sitename on the node
        :rtype: String
        """
        try:
            return self.get_fim_node().site
        except:
            return None

    def get_management_ip(self) -> str or None:
        """
        Gets the management IP on the FABRIC node.
        :return: management IP
        :rtype: String
        """
        try:
            return self.get_fim_node().management_ip
        except:
            return None

    def get_reservation_id(self) -> str or None:
        """
        Gets the reservation ID on the FABRIC node.
        :return: reservation ID on the node
        :rtype: String
        """
        try:
            return self.get_fim_node().get_property(pname='reservation_info').reservation_id
        except:
            return None

    def get_reservation_state(self) -> str or None:
        """
        Gets the reservation state on the FABRIC node.
        :return: the reservation state on the node
        :rtype: String
        """
        try:
            return self.get_fim_node().get_property(pname='reservation_info').reservation_state
        except:
            return None

    def get_error_message(self) -> str or None:
        """
        Gets the error message on the FABRIC node.
        :return: the error message on the node
        :rtype: String
        """
        try:
            return self.get_fim_node().get_property(pname='reservation_info').error_message
        except:
            return ""

    def get_interfaces(self) -> List[Interface] or None:
        """
        Gets a list of the interfaces associated with the FABRIC node.
        :return: a list of interfaces on the node
        :rtype: List[Interface]
        """
        interfaces = []
        for component in self.get_components():
            for interface in component.get_interfaces():
                interfaces.append(interface)

        return interfaces

    def get_interface(self, name: str = None, network_name: str = None) -> Interface or None:
        """
        Gets a particular interface associated with a FABRIC node.
        Accepts either the interface name or a network_name. If a network name
        is used this method will return the interface on the node that is
        connected to the network specified.
        If a name and network_name are both used, the interface name will
        take precedence.
        :param name: interface name to search for
        :type name: str
        :param network_name: network name to search for
        :type name: str
        :raise Exception: if interface is not found
        :return: an interface on the node
        :rtype: Interface
        """
        if name is not None:
            for component in self.get_components():
                for interface in component.get_interfaces():
                    if interface.get_name() == name:
                        return interface
        elif network_name is not None:
            for interface in self.get_interfaces():
                if interface is not None and interface.get_network() is not None and \
                        interface.get_network().get_name() == network_name:
                    return interface

        raise Exception("Interface not found: {}".format(name))

    def get_username(self) -> str:
        """
        Gets the username on this fablib node.
        :return: the username on this node
        :rtype: String
        """
        return self.username

    def get_public_key(self) -> str:
        """
        Gets the public key on fablib node.
        Important! Slice key management is underdevelopment and this
        functionality will likely change going forward.
        :return: the public key on the node
        :rtype: String
        """
        return self.get_slice().get_slice_public_key()

    def get_public_key_file(self) -> str:
        """
        Gets the public key file path on the fablib node.
        Important! Slice key management is underdevelopment and this
        functionality will likely change going forward.
        :return: the public key path
        :rtype: String
        """
        return self.get_slice().get_slice_public_key_file()

    def get_private_key(self) -> str:
        """
        Gets the private key on the fablib node.
        Important! Slice key management is underdevelopment and this
        functionality will likely change going forward.
        :return: the private key on the node
        :rtype: String
        """
        return self.get_slice().get_slice_private_key()

    def get_private_key_file(self) -> str:
        """
        Gets the private key file path on the fablib slice.
        Important! Slice key management is underdevelopment and this
        functionality will likely change going forward.
        :return: the private key path
        :rtype: String
        """
        return self.get_slice().get_slice_private_key_file()

    def get_private_key_passphrase(self) -> str:
        """
        Gets the private key passphrase on the FABLIB slice.
        Important! Slice key management is underdevelopment and this
        functionality will likely change going forward.
        :return: the private key passphrase
        :rtype: String
        """
        return self.get_slice().get_private_key_passphrase()

    def add_component(self, model: str = None, name: str = None) -> Component:
        """
        Creates a new FABRIC component using this fablib node.
        Example model include:
        - NIC_Basic: A single port 100 Gbps SR-IOV Virtual Function on a Mellanox ConnectX-6
        - NIC_ConnectX_5: A dual port 25 Gbps Mellanox ConnectX-5
        - NIC_ConnectX_6: A dual port 100 Gbps Mellanox ConnectX-6
        - NVME_P4510: NVMe Storage Device
        - GPU_TeslaT4: Tesla T4 GPU
        - GPU_RTX6000: RTX6000 GPU
        :param model: the name of the component model to add
        :type model: String
        :param name: the name of the new component
        :type name: String
        :return: the new component
        :rtype: Component
        """
        return Component.new_component(node=self, model=model, name=name)

    def get_components(self) -> List[Component]:
        """
        Gets a list of components associated with this node.
        :return: a list of components on this node
        :rtype: List[Component]
        """
        return_components = []
        for component_name, component in self.get_fim_node().components.items():
            # return_components.append(Component(self,component))
            return_components.append(Component(self, component))

        return return_components

    def get_component(self, name: str) -> Component:
        """
        Gets a particular component associated with this node.
        :param name: the name of the component to search for
        :type name: String
        :raise Exception: if component not found by name
        :return: the component on the FABRIC node
        :rtype: Component
        """
        try:
            name = Component.calculate_name(node=self, name=name)
            return Component(self, self.get_fim_node().components[name])
        except Exception as e:
            logging.error(e, exc_info=True)
            raise Exception(f"Component not found: {name}")

    def get_ssh_command(self) -> str:
        """
        Gets a SSH command used to access this node node from a terminal.
        :return: the SSH command to access this node
        :rtype: str
        """
        return 'ssh -i {} -J {}@{} {}@{}'.format(self.get_private_key_file(),
                                           fablib.get_bastion_username(),
                                           fablib.get_bastion_public_addr(),
                                           self.get_username(),
                                           self.get_management_ip())

    def validIPAddress(self, IP: str) -> str:
        """
        Checks if the IP string is a valid IP address.
        :param IP: the IP string to check
        :type IP: String
        :return: the type of IP address the IP string is, or 'Invalid'
        :rtype: String
        """
        try:
            return "IPv4" if type(ip_address(IP)) is IPv4Address else "IPv6"
        except ValueError:
            return "Invalid"

    def __get_paramiko_key(self, private_key_file: str = None, get_private_key_passphrase: str = None) -> paramiko.PKey:
        #TODO: This is a bit of a hack and should probably test he keys for their types
        # rather than relying on execptions
        if get_private_key_passphrase:
            try:
                return paramiko.RSAKey.from_private_key_file(self.get_private_key_file(),  password=self.get_private_key_passphrase())
            except:
                pass

            try:
                return paramiko.ecdsakey.ECDSAKey.from_private_key_file(self.get_private_key_file(),  password=self.get_private_key_passphrase())
            except:
                pass
        else:
            try:
                return paramiko.RSAKey.from_private_key_file(self.get_private_key_file())
            except:
                pass

            try:
                return paramiko.ecdsakey.ECDSAKey.from_private_key_file(self.get_private_key_file())
            except:
                pass

        raise Exception(f"ssh key invalid: FABRIC requires RSA or ECDSA keys")

    def execute(self, command: str, retry: int = 3, retry_interval: int = 10):
        """
        Runs a command on the FABRIC node.
        :param command: the command to run
        :type command: str
        :param retry: the number of times to retry SSH upon failure
        :type retry: int
        :param retry_interval: the number of seconds to wait before retrying SSH upon failure
        :type retry_interval: int
        :raise Exception: if management IP is invalid
        """
        import logging

        logging.debug(f"execute node: {self.get_name()}, management_ip: {self.get_management_ip()}, command: {command}")

        if fablib.get_log_level() == logging.DEBUG:
            start = time.time()

        # Get and test src and management_ips
        management_ip = str(self.get_fim_node().get_property(pname='management_ip'))
        if self.validIPAddress(management_ip) == 'IPv4':
            src_addr = (fablib.get_bastion_private_ipv4_addr(), 22)
        elif self.validIPAddress(management_ip) == 'IPv6':
            src_addr = (fablib.get_bastion_private_ipv6_addr(), 22)
        else:
            raise Exception(f"node.execute: Management IP Invalid: {management_ip}")
        dest_addr = (management_ip, 22)

        for attempt in range(retry):
            try:
                key = self.__get_paramiko_key(private_key_file=self.get_private_key_file(),
                                              get_private_key_passphrase=self.get_private_key_file())
                bastion = paramiko.SSHClient()
                bastion.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                bastion.connect(fablib.get_bastion_public_addr(), username=fablib.get_bastion_username(),
                                key_filename=fablib.get_bastion_key_filename())

                bastion_transport = bastion.get_transport()
                bastion_channel = bastion_transport.open_channel("direct-tcpip", dest_addr, src_addr)

                client = paramiko.SSHClient()
                client.load_system_host_keys()
                client.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
                client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

                client.connect(management_ip,username=self.username,pkey = key, sock=bastion_channel)

                stdin, stdout, stderr = client.exec_command('echo \"' + command +
                                                            '\" > /tmp/fabric_execute_script.sh; chmod +x /tmp/fabric_execute_script.sh; /tmp/fabric_execute_script.sh')
                rtn_stdout = str(stdout.read(),'utf-8').replace('\\n','\n')
                rtn_stderr = str(stderr.read(),'utf-8').replace('\\n','\n')

                client.close()
                bastion_channel.close()

                if fablib.get_log_level() == logging.DEBUG:
                    end = time.time()
                    logging.debug(f"Running node.execute(): command: {command}, elapsed time: {end - start} seconds")

                logging.debug(f"rtn_stdout: {rtn_stdout}")
                logging.debug(f"rtn_stderr: {rtn_stderr}")

                return rtn_stdout, rtn_stderr
            except Exception as e:
                try:
                    client.close()
                except:
                    logging.debug("Exception in client.close")
                    pass
                try:
                    bastion_channel.close()
                except:
                    logging.debug("Exception in bastion_channel.close()")
                    pass

                if attempt+1 == retry:
                    raise e

                #Fail, try again
                if fablib.get_log_level() == logging.DEBUG:
                    logging.debug(f"SSH execute fail. Slice: {self.get_slice().get_name()}, "
                                  f"Node: {self.get_name()}, trying again")
                    logging.debug(e, exc_info=True)

                time.sleep(retry_interval)
                pass

        raise Exception("ssh failed: Should not get here")

    def upload_file(self, local_file_path: str, remote_file_path: str, retry: int = 3, retry_interval: int = 10):
        """
        Upload a local file to a remote location on the node.
        :param local_file_path: the path to the file to upload
        :type local_file_path: str
        :param remote_file_path: the destination path of the file on the node
        :type remote_file_path: str
        :param retry: how many times to retry SCP upon failure
        :type retry: int
        :param retry_interval: how often to retry SCP on failure
        :type retry_interval: int
        :raise Exception: if management IP is invalid
        """
        logging.debug(f"upload node: {self.get_name()}, local_file_path: {local_file_path}")

        if fablib.get_log_level() == logging.DEBUG:
            start = time.time()

        # Get and test src and management_ips
        management_ip = str(self.get_fim_node().get_property(pname='management_ip'))
        if self.validIPAddress(management_ip) == 'IPv4':
            src_addr = (fablib.get_bastion_private_ipv4_addr(), 22)
        elif self.validIPAddress(management_ip) == 'IPv6':
            src_addr = (fablib.get_bastion_private_ipv6_addr(), 22)
        else:
            raise Exception(f"upload_file: Management IP Invalid: {management_ip}")
        dest_addr = (management_ip, 22)

        for attempt in range(retry):
            try:
                key = self.__get_paramiko_key(private_key_file=self.get_private_key_file(),
                                              get_private_key_passphrase=self.get_private_key_file())

                bastion=paramiko.SSHClient()
                bastion.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                bastion.connect(fablib.get_bastion_public_addr(), username=fablib.get_bastion_username(),
                                key_filename=fablib.get_bastion_key_filename())

                bastion_transport = bastion.get_transport()
                bastion_channel = bastion_transport.open_channel("direct-tcpip", dest_addr, src_addr)


                client = paramiko.SSHClient()
                client.load_system_host_keys()
                client.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
                client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

                client.connect(management_ip,username=self.username,pkey = key, sock=bastion_channel)

                ftp_client=client.open_sftp()
                file_attributes = ftp_client.put(local_file_path, remote_file_path)
                ftp_client.close()

                bastion_channel.close()

                if fablib.get_log_level() == logging.DEBUG:
                    end = time.time()
                    logging.debug(f"Running node.upload_file(): file: {local_file_path}, "
                                  f"elapsed time: {end - start} seconds")

                return file_attributes
            except Exception as e:
                try:
                    client.close()
                except:
                    logging.debug("Exception in client.close")
                    pass
                try:
                    bastion_channel.close()
                except:
                    logging.debug("Exception in bastion_channel.close()")
                    pass

                if attempt+1 == retry:
                    raise e

                #Fail, try again
                print(f"SCP upload fail. Slice: {self.get_slice().get_name()}, Node: {self.get_name()}, trying again")
                print(f"Fail: {e}")
                #traceback.print_exc()
                time.sleep(retry_interval)
                pass

        raise Exception("scp upload failed")

    def download_file(self, local_file_path: str, remote_file_path: str, retry: int = 3, retry_interval: int = 10):
        """
        Download a remote file from the node to a local destination.
        :param local_file_path: the destination path for the remote file
        :type local_file_path: str
        :param remote_file_path: the path to the remote file to download
        :type remote_file_path: str
        :param retry: how many times to retry SCP upon failure
        :type retry: int
        :param retry_interval: how often to retry SCP upon failure
        :type retry_interval: int
        :param verbose: indicator for verbose outpu
        :type verbose: bool
        """
        logging.debug(f"download node: {self.get_name()}, remote_file_path: {remote_file_path}")

        if fablib.get_log_level() == logging.DEBUG:
            start = time.time()

        # Get and test src and management_ips
        management_ip = str(self.get_fim_node().get_property(pname='management_ip'))
        if self.validIPAddress(management_ip) == 'IPv4':
            src_addr = (fablib.get_bastion_private_ipv4_addr(), 22)
        elif self.validIPAddress(management_ip) == 'IPv6':
            src_addr = (fablib.get_bastion_private_ipv6_addr(), 22)
        else:
            raise Exception(f"upload_file: Management IP Invalid: {management_ip}")
        dest_addr = (management_ip, 22)

        for attempt in range(retry):
            try:
                key = self.__get_paramiko_key(private_key_file=self.get_private_key_file(),
                                              get_private_key_passphrase=self.get_private_key_file())

                bastion=paramiko.SSHClient()
                bastion.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                bastion.connect(fablib.get_bastion_public_addr(), username=fablib.get_bastion_username(),
                                key_filename=fablib.get_bastion_key_filename())

                bastion_transport = bastion.get_transport()
                bastion_channel = bastion_transport.open_channel("direct-tcpip", dest_addr, src_addr)

                client = paramiko.SSHClient()
                client.load_system_host_keys()
                client.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
                client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

                client.connect(management_ip,username=self.username,pkey = key, sock=bastion_channel)

                ftp_client = client.open_sftp()
                file_attributes = ftp_client.get(remote_file_path, local_file_path)
                ftp_client.close()

                bastion_channel.close()

                if fablib.get_log_level() == logging.DEBUG:
                    end = time.time()
                    logging.debug(f"Running node.download(): file: {remote_file_path}, "
                                  f"elapsed time: {end - start} seconds")

                return file_attributes
            except Exception as e:
                try:
                    client.close()
                except:
                    logging.debug("Exception in client.close")
                    pass
                try:
                    bastion_channel.close()
                except:
                    logging.debug("Exception in bastion_channel.close()")
                    pass

                if attempt+1 == retry:
                    raise e

                #Fail, try again
                print(f"SCP download fail. Slice: {self.get_slice().get_name()}, Node: {self.get_name()}, trying again")
                print(f"Fail: {e}")
                #traceback.print_exc()
                time.sleep(retry_interval)
                pass

        raise Exception("scp download failed")

    def upload_directory(self, local_directory_path: str, remote_directory_path: str, retry: int = 3,
                         retry_interval: int = 10):
        """
        Upload a directory to remote location on the node.
        Makes a gzipped tarball of a directory and uploades it to a node. Then
        unzips and tars the directory at the remote_directory_path
        :param local_directory_path: the path to the directory to upload
        :type local_directory_path: str
        :param remote_directory_path: the destination path of the directory on the node
        :type remote_directory_path: str
        :param retry: how many times to retry SCP upon failure
        :type retry: int
        :param retry_interval: how often to retry SCP on failure
        :type retry_interval: int
        :raise Exception: if management IP is invalid
        """
        import tarfile
        import os

        logging.debug(f"upload node: {self.get_name()}, local_directory_path: {local_directory_path}")

        output_filename = local_directory_path.split('/')[-1]
        root_size = len(local_directory_path) - len(output_filename)
        temp_file = "/tmp/" + output_filename + ".tar.gz"

        with tarfile.open(temp_file, "w:gz") as tar_handle:
            for root, dirs, files in os.walk(local_directory_path):
                for file in files:
                    tar_handle.add(os.path.join(root, file), arcname = os.path.join(root, file)[root_size:])

        self.upload_file(temp_file, temp_file, retry, retry_interval)
        os.remove(temp_file)
        self.execute("mkdir -p "+remote_directory_path + "; tar -xf " + temp_file + " -C " +
                     remote_directory_path + "; rm " + temp_file, retry, retry_interval)
        return "success"

    def download_directory(self, local_directory_path: str, remote_directory_path: str, retry: int = 3,
                           retry_interval: int = 10):
        """
        Downloads a directory from remote location on the node.
        Makes a gzipped tarball of a directory and downloads it from a node. Then
        unzips and tars the directory at the local_directory_path
        :param local_directory_path: the path to the directory to upload
        :type local_directory_path: str
        :param remote_directory_path: the destination path of the directory on the node
        :type remote_directory_path: str
        :param retry: how many times to retry SCP upon failure
        :type retry: int
        :param retry_interval: how often to retry SCP on failure
        :type retry_interval: int
        :raise Exception: if management IP is invalid
        """
        import tarfile
        import os
        logging.debug(f"upload node: {self.get_name()}, local_directory_path: {local_directory_path}")

        temp_file = "/tmp/unpackingfile.tar.gz"
        self.execute("tar -czf " + temp_file + " " + remote_directory_path, retry, retry_interval)

        self.download_file(temp_file, temp_file, retry, retry_interval)
        tar_file = tarfile.open(temp_file)
        tar_file.extractall(local_directory_path)

        self.execute("rm " + temp_file, retry, retry_interval)
        os.remove(temp_file)
        return "success"

    def test_ssh(self) -> bool:
        """
        Test whether SSH is functional on the node.
        :return: true if SSH is working, false otherwise
        :rtype: bool
        """
        logging.debug(f"test_ssh: node {self.get_name()}")

        try:
            if self.get_management_ip() is None:
                logging.debug(f"Node: {self.get_name()} failed test_ssh because management_ip == None" )

            self.execute(f'echo test_ssh from {self.get_name()}', retry=1, retry_interval=10)
        except Exception as e:
            #logging.debug(f"{e}")
            logging.debug(e, exc_info=True)
            return False
        return True

    def get_management_os_interface(self) -> str or None:
        """
        Gets the name of the management interface used by the node's operating
        system.
        :return: interface name
        :rtype: String
        """
        # TODO: Add docstring after doc networking classes
        #Assumes that the default route uses the management network
        logging.debug(f"{self.get_name()}->get_management_os_interface")
        stdout, stderr = self.execute("sudo ip -j route list")
        stdout_json = json.loads(stdout)

        #print(pythonObj)
        for i in stdout_json:
            if i['dst'] == 'default':
                logging.debug(f"{self.get_name()}->get_management_os_interface: management_os_interface {i['dev']}")
                return i['dev']
        return None

    def get_dataplane_os_interfaces(self) -> List[dict]:
        """
        Gets a list of all the dataplane interface names used by the node's
        operating system.
        :return: interface names
        :rtype: List[String]
        """
        management_dev = self.get_management_os_interface()

        stdout, stderr = self.execute("sudo ip -j addr list")
        stdout_json = json.loads(stdout)
        dataplane_devs = []
        for i in stdout_json:
            if i['ifname'] != 'lo' and i['ifname'] !=  management_dev:
                dataplane_devs.append({'ifname': i['ifname'], 'mac': i['address']})

        return dataplane_devs

    def flush_all_os_interfaces(self):
        """
        Flushes the configuration of all dataplane interfaces in the node.
        """
        for iface in self.get_dataplane_os_interfaces():
            self.flush_os_interface(iface['ifname'])

    def flush_os_interface(self, os_iface: str):
        """
        Flush the configuration of an interface in the node
        :param os_iface: the name of the interface to flush
        :type os_iface: String
        """
        stdout, stderr = self.execute(f"sudo ip addr flush dev {os_iface}")
        stdout, stderr = self.execute(f"sudo ip -6 addr flush dev {os_iface}")

    def validIPAddress(self, IP: str) -> str:
        try:
            return "IPv4" if type(ip_address(IP)) is IPv4Address else "IPv6"
        except ValueError:
            return "Invalid"

    def ip_route_add(self, subnet: Union[IPv4Network, IPv6Network], gateway: Union[IPv4Address, IPv6Address]):
        """
        Add a route on the node.
        :param subnet: The destination subnet
        :type subnet:  IPv4Network or IPv6Network
        :param gateway: The next hop gateway.
        :type gateway: IPv4Address or IPv6Address
        """
        ip_command = ""
        if type(subnet) == IPv6Network:
            ip_command = "sudo ip -6"
        elif type(subnet) == IPv4Network:
            ip_command = "sudo ip"

        try:
            self.execute(f"{ip_command} route add {subnet} via {gateway}")
        except Exception as e:
            logging.warning(f"Failed to add route: {e}")
            raise e

    def ip_route_del(self, subnet: Union[IPv4Network, IPv6Network], gateway: Union[IPv4Address, IPv6Address]):
        """
        Delete a route on the node.
        :param subnet: The destination subnet
        :type subnet:  IPv4Network or IPv6Network
        :param gateway: The next hop gateway.
        :type gateway: IPv4Address or IPv6Address
        """
        ip_command = ""
        if type(subnet) == IPv6Network:
            ip_command = "sudo ip -6"
        elif type(subnet) == IPv4Network:
            ip_command = "sudo ip"

        try:
            self.execute(f"{ip_command} route del {subnet} via {gateway}")
        except Exception as e:
            logging.warning(f"Failed to del route: {e}")
            raise e

    def ip_addr_add(self, addr: Union[IPv4Address, IPv6Address], subnet: Union[IPv4Network, IPv6Network],
                    interface: Interface):
        """
        Add an IP to an interface on the node.
        :param addr: IP address
        :type addr:  IPv4Address or IPv6Address
        :param subnet: subnet.
        :type subnet: IPv4Network or IPv6Network
        :param interface: the FABlib interface.
        :type interface: Interface
        """
        ip_command = ""
        if type(subnet) == IPv6Network:
            ip_command = "sudo ip -6"
        elif type(subnet) == IPv4Network:
            ip_command = "sudo ip"

        try:
            self.execute(f"{ip_command} addr add {addr}/{subnet.prefixlen} dev {interface.get_os_interface()} ")
        except Exception as e:
            logging.warning(f"Failed to add addr: {e}")
            raise e

    def ip_addr_del(self, addr: Union[IPv4Address, IPv6Address], subnet: Union[IPv4Network, IPv6Network],
                    interface: Interface):
        """
        Delete an IP to an interface on the node.
        :param addr: IP address
        :type addr:  IPv4Address or IPv6Address
        :param subnet: subnet.
        :type subnet: IPv4Network or IPv6Network
        :param interface: the FABlib interface.
        :type interface: Interface
        """
        ip_command = ""
        if type(subnet) == IPv6Network:
            ip_command = "sudo ip -6"
        elif type(subnet) == IPv4Network:
            ip_command = "sudo ip"

        try:
            self.execute(f"{ip_command} addr del {addr}/{subnet.prefixlen} dev {interface.get_os_interface()} ")
        except Exception as e:
            logging.warning(f"Failed to del addr: {e}")
            raise e

    def ip_link_up(self, subnet: Union[IPv4Network, IPv6Network], interface: Interface):
        """
        Bring up a link on an interface on the node.
        :param subnet: subnet.
        :type subnet: IPv4Network or IPv6Network
        :param interface: the FABlib interface.
        :type interface: Interface
        """
        if type(subnet) == IPv6Network:
            ip_command = "sudo ip -6"
        elif type(subnet) == IPv4Network:
            ip_command = "sudo ip"

        try:
            self.execute(f"{ip_command} link set dev {interface.get_os_interface()} up")
        except Exception as e:
            logging.warning(f"Failed to up link: {e}")
            raise e

    def ip_link_down(self, subnet: Union[IPv4Network, IPv6Network], interface: Interface):
        """
        Bring down a link on an interface on the node.
        :param subnet: subnet.
        :type subnet: IPv4Network or IPv6Network
        :param interface: the FABlib interface.
        :type interface: Interface
        """
        if type(subnet) == IPv6Network:
            ip_command = "sudo ip -6"
        elif type(subnet) == IPv4Network:
            ip_command = "sudo ip"

        try:
            self.execute(f"{ip_command} link set dev {interface.get_os_interface()} down")
        except Exception as e:
            logging.warning(f"Failed to up link: {e}")
            raise e

    def set_ip_os_interface(self, os_iface: str = None, vlan: str = None, ip: str = None, cidr: str = None,
                            mtu: str = None):
        """
        Depricated
        """
        # TODO: Add docstring after doc networking classes
        if cidr: cidr=str(cidr)
        if mtu: mtu=str(mtu)

        if self.validIPAddress(ip) == "IPv4":
            ip_command = "sudo ip"
        elif self.validIPAddress(ip) == "IPv6":
            ip_command = "sudo ip -6"
        else:
            raise Exception(f"Invalid IP {ip}. IP must be vaild IPv4 or IPv6 string.")

        #Bring up base iface
        logging.debug(f"{self.get_name()}->set_ip_os_interface: os_iface {os_iface}, vlan {vlan}, ip {ip}, cidr {cidr}, mtu {mtu}")
        command = f'{ip_command} link set dev {os_iface} up'

        if mtu != None:
            command += f" mtu {mtu}"
        stdout, stderr = self.execute(command)

        #config vlan iface
        if vlan != None:
            #create vlan iface
            command = f'{ip_command} link add link {os_iface} name {os_iface}.{vlan} type vlan id {vlan}'
            stdout, stderr = self.execute(command)

            #bring up vlan iface
            os_iface = f"{os_iface}.{vlan}"
            command = f'{ip_command} link set dev {os_iface} up'
            if mtu != None:
                command += f" mtu {mtu}"
            stdout, stderr = self.execute(command)

        if ip != None and cidr != None:
            #Set ip
            command = f"{ip_command} addr add {ip}/{cidr} dev {os_iface}"
            stdout, stderr = self.execute(command)

        stdout, stderr = self.execute(command)

    def clear_all_ifaces(self):
        """
        Flush all interfaces and delete VLAN os interfaces
        """
        # TODO: Add docstring after doc networking classes
        self.remove_all_vlan_os_interfaces()
        self.flush_all_os_interfaces()

    def remove_all_vlan_os_interfaces(self):
        """
        Delete all VLAN os interfaces
        """
        # TODO: Add docstring after doc networking classes
        management_os_iface = self.get_management_os_interface()

        stdout, stderr = self.execute("sudo ip -j addr list")
        stdout_json = json.loads(stdout)
        dataplane_devs = []
        for i in stdout_json:
            if i['ifname'] == management_os_iface or i['ifname'] == 'lo':
                stdout_json.remove(i)
                continue

            #If iface is vlan linked to base iface
            if 'link' in i.keys():
                self.remove_vlan_os_interface(os_iface=i['ifname'])

    def get_interface_map(self):
        """
        Not intended for API use.
        """
        # TODO: Add docstring after doc networking classes
        #data = {}
        #Get interface data
        logging.debug(f"get_interface_map: node {self.get_name()}")

        interfaces = {}
        for i in self.get_interfaces():
            logging.debug(f"get_interface_map: i: {i}")

            #print(f"interface: {i.get_name()}")
            #print(f"os_interface: {i.get_physical_os_interface()}")
            if i.get_network() != None:
                logging.debug(f"i.get_network().get_name(): {i.get_network().get_name()}")
                network_name = i.get_network().get_name()
                #print(f"network: {i.get_network().get_name()}")
            else:
                logging.debug(f"i.get_network(): None")
                network_name = None
                #print(f"network: None")

            interfaces[i.get_name()] =  { 'network':  network_name,
                        'os_interface':  i.get_physical_os_interface() }
        return interfaces

    def save_data(self):
        """
        Not intended for API use.
        """
        # TODO: Add docstring after doc networking classes
        logging.debug(f"save_data: node {self.get_name()}")

        interfaces = {}
        for i in self.get_interfaces():
            #print(f"interface: {i.get_name()}")
            #print(f"os_interface: {i.get_physical_os_interface()}")
            if i.get_network() != None:
                network_name = i.get_network().get_name()
                #print(f"network: {i.get_network().get_name()}")
            else:
                network_name = None
                #print(f"network: None")

            interfaces[i.get_name()] =  { 'network':  network_name,
                         'os_interface':  i.get_physical_os_interface() }

        with open(f'/tmp/fablib/fabric_data/{self.get_name()}.json', 'w') as outfile:
            json.dump(interfaces, outfile)

        #print(f"interfaces: {json.dumps(interfaces).replace('\"','\\"')}")

        self.upload_file(f'/tmp/fablib/fabric_data/{self.get_name()}.json', f'{self.get_name()}.json')

    def load_data(self):
        """
        Not intended for API use.
        """
        # TODO: Add docstring after doc networking classes
        logging.debug(f"load_data: node {self.get_name()}")

        try:
            self.download_file(f'/tmp/fablib/fabric_data/{self.get_name()}.json',f'{self.get_name()}.json')

            interfaces=""
            with open(f'/tmp/fablib/fabric_data/{self.get_name()}.json', 'r') as infile:
                interfaces = json.load(infile)

            interface_map = self.get_slice().network_iface_map #= self.get_slice().get_interface_map()
            #print(f"interfaces {interfaces}")
            for interface_name, net_map in interfaces.items():
                logging.debug(f"interface_name: {interface_name}, {net_map}")
                if net_map['network'] != None:
                    interface_map[net_map['network']][self.get_name()] = net_map['os_interface']

            self.get_slice().network_iface_map = interface_map
            logging.debug(f"{self.get_slice().network_iface_map}")
        except Exception as e:
            logging.error(f"load data fail: {e}")
            logging.error(e, exc_info=True)
            raise e

    def remove_vlan_os_interface(self, os_iface: str = None):
        """
        Remove one VLAN OS interface
        """
        # TODO: Add docstring after doc networking classes
        command = f"sudo ip -j addr show {os_iface}"
        stdout, stderr = self.execute(command)
        try:
            [stdout_json] = json.loads(stdout)
        except Exception as e:
            print(f"os_iface: {os_iface}, stdout: {stdout}, stderr: {stderr}")
            raise e

        link = stdout_json['link']

        command = f"sudo ip link del link {link} name {os_iface}"
        stdout, stderr = self.execute(command)

    def add_vlan_os_interface(self, os_iface: str = None, vlan: str = None, ip: str = None, cidr: str = None,
                              mtu: str = None, interface: str = None):
        """
        Depricated
        """
        # TODO: Add docstring after doc networking classes

        if vlan: vlan=str(vlan)
        if cidr: cidr=str(cidr)
        if mtu: mtu=str(mtu)

        try:
            gateway = None
            if interface.get_network().get_layer() == NSLayer.L3:
                if interface.get_network().get_type() == ServiceType.FABNetv6:
                    ip_command = "sudo ip -6"
                elif interface.get_network().get_type() == ServiceType.FABNetv4:
                    ip_command = "sudo ip"
            else:
                ip_command = "sudo ip"
        except Exception as e:
            logging.warning(f"Failed to get network layer and/or type: {e}")
            ip_command = "sudo ip"


        #if interface. == "IPv4":
        #    ip_command = "sudo ip"
        #elif self.validIPAddress(ip) == "IPv6":
        #    ip_command = "sudo ip -6"
        #else:
        #    logging.debug(f"Invalid IP {ip}. IP must be vaild IPv4 or IPv6 string. Config VLAN interface only.")

        command = f'{ip_command} link add link {os_iface} name {os_iface}.{vlan} type vlan id {vlan}'

        stdout, stderr = self.execute(command)
        command = f'{ip_command} link set dev {os_iface}.{vlan} up'
        stdout, stderr = self.execute(command)

        if ip != None and cidr != None:
            self.set_ip_os_interface(os_iface=f"{os_iface}.{vlan}", ip=ip, cidr=cidr, mtu=mtu)

    def ping_test(self, dst_ip: str) -> bool:
        """
        Test a ping from the node to a destination IP
        :param dst_ip: destination IP String.
        :type dst_ip: String
        """
        # TODO: Add docstring after doc networking classes
        logging.debug(f"ping_test: node {self.get_name()}")

        command = f'ping -c 1 {dst_ip}  2>&1 > /dev/null && echo Success'
        stdout, stderr = self.execute(command)
        if stdout.replace("\n","") == 'Success':
            return True
        else:
            return False
